/*!
        @file    $Id:: action_F_Staggered_eo.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-07-12 14:43:23 #$

        @version $LastChangedRevision: 929 $
*/

#include "action_F_Staggered_eo.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("rho_uniform", 0.0);
    param.Register_int("number_of_smearing", 0);

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Action.F_Staggered_eo", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Action_F_Staggered_eo::Parameters_Action_F_Staggered_eo() { append_entry(*this); }
//- end

//====================================================================
void Action_F_Staggered_eo::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double rho;
  int    Nsmear;

  int err = 0;
  err += params.fetch_double("rho_uniform", rho);
  err += params.fetch_int("number_of_smearing", Nsmear);

  if (err) {
    vout.crucial(m_vl, "Action_F_Staggered_eo: fetch error, input parameter not found.\n");
    abort();
  }


  set_parameters(Nsmear, rho);
}


//====================================================================
void Action_F_Staggered_eo::set_parameters(const Parameters_Fopr_Staggered_eo& params)
{
  double rho    = params.get_double("rho_uniform");
  int    Nsmear = params.get_int("number_of_smearing");

  set_parameters(Nsmear, rho);
}


//====================================================================
void Action_F_Staggered_eo::set_parameters(int Nsmear, double rho)
{
  //- print input parameters
  vout.general(m_vl, "Action_F_Staggered_eo:\n");
  vout.general(m_vl, "  Nsmear = %4d\n", Nsmear);
  vout.general(m_vl, "  rho    = %8.4f\n", rho);

  //- range check
  // NB. Nsmear,rho == 0 is allowed.

  //- store values
  m_Nsmear = Nsmear;
  m_rho    = rho;

  //- post-process
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();
  int NinG = 2 * Nc * Nc;

  m_force.reset(NinG, Nvol, Ndim);

  // m_smear->set_parameters(m_rho);
  Parameters *params_smear = ParametersFactory::New("Smear");
  params_smear->set_double("rho_uniform", rho);
  m_smear->set_parameters(*params_smear);

  // m_force_smear->set_parameters(m_rho);
  Parameters *params_force_smear = ParametersFactory::New("ForceSmear");
  params_force_smear->set_double("rho_uniform", rho);
  m_force_smear->set_parameters(*params_force_smear);

  m_Usmear.resize(m_Nsmear);
  for (int i_smear = 0; i_smear < m_Nsmear; ++i_smear) {
    m_Usmear[i_smear].reset(NinG, Nvol, Ndim);
  }

  //- link variable update flag
  m_status_linkv = 0;

  delete params_smear;
  delete params_force_smear;
}


//====================================================================
double Action_F_Staggered_eo::langevin(RandomNumbers *rand)
{
  int Nc       = CommonParameters::Nc();
  int Nd       = CommonParameters::Nd();
  int Lvol     = CommonParameters::Lvol();
  int Nvol     = CommonParameters::Nvol();
  int Ndim     = CommonParameters::Ndim();
  int Nvol2    = Nvol / 2;
  int size_psf = Lvol * Nc * Nd;

  Field_F_1spinor xi(Nvol, 1);

  rand->gauss_lex_global(xi);

  m_psf.reset(2 * Nc, Nvol2, 1);

  Field_G Usmear(Nvol, Ndim), Uprev(Nvol, Ndim);

  if (m_Nsmear == 0) {
    Usmear = *m_U;
  }

  for (int i_smear = 0; i_smear < m_Nsmear; ++i_smear) {
    if (i_smear == 0) {
      Uprev = *m_U;
    } else {
      Uprev = (Field_G)m_Usmear[i_smear - 1];
    }
    m_smear->smear(Usmear, Uprev);
    m_Usmear[i_smear] = (Field)Usmear;
  }

  Field_G Ueo(Nvol, Ndim);
  m_index_eo.convertField(Ueo, Usmear);

  m_fopr->set_config(&Ueo);

  Field_F_1spinor xi_e(Nvol2, 1), xi_o(Nvol2, 1);
  m_index_eo.convertField(xi_e, xi, 0);
  m_index_eo.convertField(xi_o, xi, 1);

  m_psf = (Field)xi_e;
  m_fopr->set_mode("Meo");
  m_psf -= m_fopr->mult(xi_o);
  //  m_fopr->fopr_normalize(m_psf);

  double H_psf = calcH();

  m_status_linkv = 0;

  return H_psf;
}


//====================================================================
double Action_F_Staggered_eo::calcH()
{
  int Nc       = CommonParameters::Nc();
  int Lvol     = CommonParameters::Lvol();
  int Nvol     = CommonParameters::Nvol();
  int Ndim     = CommonParameters::Ndim();
  int size_psf = (Lvol / 2) * Nc;

  int    Niter     = 5000;
  double Stop_cond = 1.0e-24;
  int    Nconv;
  double diff;

  Field v1(m_psf);

  assert(v1.nvol() == Nvol / 2);

  Field_G Usmear(Nvol, Ndim), Uprev(Nvol, Ndim);

  if (m_Nsmear == 0) {
    Usmear = *m_U;
  }

  for (int i_smear = 0; i_smear < m_Nsmear; ++i_smear) {
    if (i_smear == 0) {
      Uprev = *m_U;
    } else {
      Uprev = (Field_G)m_Usmear[i_smear - 1];
    }
    m_smear->smear(Usmear, Uprev);
    m_Usmear[i_smear] = (Field)Usmear;
  }

  Field_G Ueo(Nvol, Ndim);
  m_index_eo.convertField(Ueo, Usmear);

  m_fopr->set_config(&Ueo);
  // This must be replaced with "DdagD".
  m_fopr->set_mode("Dee");

  m_solver->set_parameters(Niter, Stop_cond);
  m_solver->solve(v1, m_psf, Nconv, diff);

  vout.general(m_vl, "  Nconv = %d\n", Nconv);
  vout.general(m_vl, "  diff  = %.8e\n", diff);

  double H_psf = v1 * m_psf;

  vout.general(m_vl, "H_Fstaggered_eo = %18.8f\n", H_psf);
  vout.general(m_vl, "H_F/dof    = %18.8f\n", H_psf / size_psf);


  return H_psf;
}


//====================================================================
const Field Action_F_Staggered_eo::force()
{
  if (m_status_linkv == 0) {
    int    Niter     = 500;
    double Stop_cond = 1.0e-24;
    int    Nconv;
    double diff;

    int Nin  = m_U->nin();
    int Nvol = m_U->nvol();
    int Nex  = m_U->nex();
    int Nc   = CommonParameters::Nc();
    int Ndim = CommonParameters::Ndim();

    Field_F_1spinor eta(Nvol / 2, 1);

    Field_G Usmear(Nvol, Ndim), Uprev(Nvol, Ndim);

    vout.general(m_vl, "  Action_F_Staggered_eo:\n");

    if (m_Nsmear == 0) {
      Usmear = *m_U;
    }

    for (int i_smear = 0; i_smear < m_Nsmear; ++i_smear) {
      if (i_smear == 0) {
        Uprev = *m_U;
      } else {
        Uprev = (Field_G)m_Usmear[i_smear - 1];
      }
      m_smear->smear(Usmear, Uprev);
      m_Usmear[i_smear] = (Field)Usmear;
    }


    // fermion inversion for smeared gauge field

    Field_G Ueo(Nvol, Ndim);
    m_index_eo.convertField(Ueo, Usmear);

    m_fopr->set_config(&Ueo);
    // This must be replaced with "DdagD".
    m_fopr->set_mode("Dee");

    m_solver->set_parameters(Niter, Stop_cond);
    m_solver->solve(eta, m_psf, Nconv, diff);

    vout.general(m_vl, "    Solver: Nconv = %6d  diff  = %12.6e\n", Nconv, diff);


    // force of smeared fermion operator
    m_fopr_force->set_config(&Ueo);

    Field_G  force(Nvol, Nex);
    Field_G  force1(Nvol, Nex);
    Mat_SU_N ut(Nc);

    if (m_Nsmear == 0) {
      force = m_fopr_force->force_core(eta);
    } else {
      force1 = m_fopr_force->force_udiv(eta);

      for (int i_smear = m_Nsmear - 1; i_smear > 0; --i_smear) {
        Usmear = (Field_G)m_Usmear[i_smear - 1];
        force  = m_force_smear->force_udiv(force1, Usmear);
        force1 = force;
      }

      Usmear = *m_U;
      //    force = m_force_smear->force_core(force1,Usmear);
      force  = m_force_smear->force_udiv(force1, Usmear);
      force1 = force;

      for (int mu = 0; mu < Ndim; ++mu) {
        for (int site = 0; site < Nvol; ++site) {
          ut = Usmear.mat(site, mu) * force1.mat(site, mu);
          ut.at();
          ut *= -2.0;
          force.set_mat(site, mu, ut);
        }
      }
    }

    m_force = (Field)force;
    ++m_status_linkv;

    double Fave, Fmax, Fdev;
    m_force.stat(Fave, Fmax, Fdev);
    vout.general(m_vl, "    Fstaggered_eo_ave = %12.6f  Fstaggered_eo_max = %12.6f  Fstaggered_eo_dev = %12.6f\n",
                 Fave, Fmax, Fdev);

    return m_force;
  } else {
    vout.general(m_vl, "  Action_F_Staggered_eo returns previous force.\n");

    return m_force;
  }
}


//====================================================================
//============================================================END=====
