/*!
        @file    $Id:: fopr_Overlap_5d.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-07-18 11:55:46 #$

        @version $LastChangedRevision: 934 $
*/

#include "fopr_Overlap_5d.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using std::valarray;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("quark_mass", 0.0);
    param.Register_double("domain_wall_height", 0.0);
    param.Register_int("number_of_poles", 0);
    param.Register_double("lower_bound", 0.0);
    param.Register_double("upper_bound", 0.0);
    param.Register_int("maximum_number_of_iteration", 0);
    param.Register_double("convergence_criterion_squared", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Fopr.Overlap_5d", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Fopr_Overlap_5d::Parameters_Fopr_Overlap_5d() { append_entry(*this); }
//- end

//====================================================================
void Fopr_Overlap_5d::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        mq, M0;
  int           Np;
  double        x_min, x_max;
  int           Niter_ms;
  double        Stop_cond_ms;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("quark_mass", mq);
  err += params.fetch_double("domain_wall_height", M0);
  err += params.fetch_int("number_of_poles", Np);
  err += params.fetch_double("lower_bound", x_min);
  err += params.fetch_double("upper_bound", x_max);
  err += params.fetch_int("maximum_number_of_iteration", Niter_ms);
  err += params.fetch_double("convergence_criterion_squared", Stop_cond_ms);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "Fopr_Overlap_5d: fetch error, input parameter not found.\n");
    abort();
  }

  set_parameters(mq, M0, Np, x_min, x_max, Niter_ms, Stop_cond_ms, bc);
}


//====================================================================
void Fopr_Overlap_5d::set_parameters(const double mq, const double M0,
                                     const int Np, const double x_min, const double x_max,
                                     const int Niter_ms, const double Stop_cond_ms,
                                     const std::valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Parameters of 5D overlap fermion operator:\n");
  vout.general(m_vl, "  mq    = %8.4f\n", mq);
  vout.general(m_vl, "  M0    = %8.4f\n", M0);
  vout.general(m_vl, "  Np    = %4d\n", Np);
  vout.general(m_vl, "  x_min = %10.6f\n", x_min);
  vout.general(m_vl, "  x_max = %10.6f\n", x_max);
  vout.general(m_vl, "  Niter_ms     = %8d\n", Niter_ms);
  vout.general(m_vl, "  Stop_cond_ms = %10.4e\n", Stop_cond_ms);
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", mu, bc[mu]);
  }

  //- range check
  int err = 0;
  err += ParameterCheck::non_zero(mq);
  err += ParameterCheck::non_zero(M0);
  err += ParameterCheck::non_zero(Np);
  // NB. x_min,x_max == 0 is allowed.
  err += ParameterCheck::non_zero(Niter_ms);
  err += ParameterCheck::square_non_zero(Stop_cond_ms);

  if (err) {
    vout.crucial(m_vl, "Fopr_Overlap: parameter range check failed.\n");
    abort();
  }

  assert(bc.size() == Ndim);

  //- store values
  m_mq = mq;
  m_M0 = M0;

  m_Np    = Np;
  m_x_min = x_min;
  m_x_max = x_max;

  m_Niter_ms     = Niter_ms;
  m_Stop_cond_ms = Stop_cond_ms;

  m_boundary.resize(Ndim);
  for (int mu = 0; mu < Ndim; ++mu) {
    m_boundary[mu] = bc[mu];
  }

  //- propagate parameters
  //- Zolotarev coefficients and shift values
  m_sigma.resize(m_Np);
  m_cl.resize(2 * m_Np);
  m_bl.resize(m_Np);

  //- Zolotarev coefficient defined
  double              bmax = m_x_max / m_x_min;
  Math_Sign_Zolotarev sign_func(m_Np, bmax);
  sign_func.get_sign_parameters(m_cl, m_bl);

  for (int i = 0; i < m_Np; i++) {
    m_sigma[i] = m_cl[2 * i] * m_x_min * m_x_min;
  }

  for (int i = 0; i < m_Np; i++) {
    vout.general(m_vl, " %3d %12.4e %12.4e %12.4e\n",
                 i, m_cl[i], m_cl[i + m_Np], m_bl[i]);
  }

  m_p_sqrt.resize(m_Np);
  m_q_sqrt.resize(m_Np);

  double b_sum = 0.0;
  for (int ip = 0; ip < m_Np; ++ip) {
    int    ik    = m_Np - ip - 1;
    double p_tmp = (m_M0 - 0.5 * m_mq) * m_bl[ip] *
                   (m_cl[2 * m_Np - 1] - m_cl[2 * ip]) * m_x_min;
    m_p_sqrt[ik] = sqrt(p_tmp);
    m_q_sqrt[ik] = sqrt(m_cl[2 * ip] * m_x_min * m_x_min);
    b_sum        = b_sum + m_bl[ip];
  }

  m_p0_parameter = (m_M0 - 0.5 * m_mq) * b_sum / m_x_min;
  m_R_parameter  = m_M0 + 0.5 * m_mq;
  m_h            = 1.0;

  m_rl.resize(m_Np);
  m_sl.resize(m_Np);
  for (int ik = 0; ik < m_Np; ++ik) {
    m_rl[ik] = -m_q_sqrt[ik];
    m_sl[ik] = -m_p_sqrt[ik] / (1.0 + m_q_sqrt[ik] * m_q_sqrt[ik]);
  }

  m_u0 = 0.0;
  for (int ik = 0; ik < m_Np; ++ik) {
    m_u0 = m_u0 - m_sl[ik] * m_p_sqrt[ik];
  }
}


//====================================================================
void Fopr_Overlap_5d::set_lowmodes(int Nsbt, valarray<double> *ev,
                                   valarray<Field> *vk)
{
  m_Nsbt = Nsbt;

  vout.general(m_vl, "  Nsbt = %d", Nsbt);

  if (m_Nsbt > 0) {
    m_ev.resize(m_Nsbt);
    m_vk.resize(2 * m_Nsbt);

    int Nin   = (*vk)[0].nin();
    int Nvol  = (*vk)[0].nvol();
    int Nvol2 = Nvol / 2;

    for (int k = 0; k < 2 * m_Nsbt; ++k) {
      m_vk[k].reset(Nin, Nvol2, 1);
    }

    Field vt(Nin, Nvol, 1);
    Field vt_eo(Nin, Nvol2, 1);

    for (int k = 0; k < m_Nsbt; ++k) {
      m_ev[k] = (*ev)[k];
      vt      = (*vk)[k];
      m_index_eo.convertField(m_vk[k], vt, 0);
      m_index_eo.convertField(m_vk[k + m_Nsbt], vt, 1);
    }
  }

  //- setting parameter for low-mode subtraction
  m_prf.resize(m_Nsbt);
  for (int k = 0; k < m_Nsbt; ++k) {
    double sign_ev = m_ev[k] / fabs(m_ev[k]);
    m_prf[k] = (m_M0 - 0.5 * m_mq) * sign_ev
               - m_p0_parameter * m_ev[k];
  }

  int Nsbt2 = 2 * m_Nsbt;

  m_u0c_e.resize(Nsbt2 * Nsbt2);
  m_u0cinv_e.resize(Nsbt2 * Nsbt2);
  m_u0c_o.resize(Nsbt2 * Nsbt2);
  m_u0cinv_o.resize(Nsbt2 * Nsbt2);

  if (m_Nsbt > 0) {
    Calc_Coeff_u0inv();
  }
}


//====================================================================
const Field Fopr_Overlap_5d::DdagD_eo(const Field& f)
{
  Field v(f);
  Field w(f);

  v = DD_5d_eo(f, 1);
  w = DD_5d_eo(v, -1);

  return w;
}


//====================================================================
const Field Fopr_Overlap_5d::DD_5d_eo(const Field& w, const int jd)
{
  int   Nin  = w.nin();
  int   Nvol = w.nvol();
  int   Nex  = w.nex();
  Field t(Nin, Nvol, Nex), v(Nin, Nvol, Nex);

  if (jd == 1) {
    Mopr_5d_eo(t, w, 1);
    LUprecond(v, t, 1);
    Mopr_5d_eo(t, v, 0);
    LUprecond(v, t, 0);
    v = -v + w;
  } else if (jd == -1) {
    LUprecond(t, w, 0);
    Mopr_5d_eo(v, t, 1);
    LUprecond(t, v, 1);
    Mopr_5d_eo(v, t, 0);
    v = -v + w;
  } else {
    vout.crucial(m_vl, "Fopr_overlap_5d: illegal jd.\n");
    abort();
  }

  return v;
}


//====================================================================
void Fopr_Overlap_5d::Mopr_5d_eo(Field& v, const Field& w,
                                 const int ieo)
{
  //       ieo = 1: M_eo, 2: M_oe

  int ieo1 = ieo;
  int ieo2 = 1 - ieo;

  int Nin   = w.nin();
  int Nvol2 = w.nvol();
  int Nex   = w.nex();

  Field z(Nin, Nvol2, 1);
  Field z1(Nin, Nvol2, 1);
  Field z2(Nin, Nvol2, 1);

  Field w1(Nin, Nvol2, 1);
  Field v1(Nin, Nvol2, 1);

  w1.setpart_ex(0, w, 2 * m_Np);
  Proj_H_eo(ieo1, ieo2, z1, w1);

  z2 = 0.0;

  for (int j = 0; j < m_Np; ++j) {
    w1.setpart_ex(0, w, 2 * j);
    v1 = (Field)m_fopr_w->Meo_gm5(w1, ieo);
    v.setpart_ex(2 * j, v1, 0);

    w1.setpart_ex(0, w, 2 * j + 1);
    z2 += m_p_sqrt[j] * w1;

    v1  = (Field)m_fopr_w->Meo_gm5(w1, ieo);
    v1 *= -1.0;
    v1 += m_p_sqrt[j] * z1;
    v.setpart_ex(2 * j + 1, v1, 0);
  }

  Proj_H_eo(ieo1, ieo2, v1, z2);
  w1.setpart_ex(0, w, 2 * m_Np);
  Proj_L_mult_eo(ieo1, ieo2, z1, w1);
  v1 += z1;
  z1  = (Field)m_fopr_w->Meo_gm5(w1, ieo);
  v1 += m_p0_parameter * z1;
  v.setpart_ex(2 * m_Np, v1, 0);
}


//====================================================================
void Fopr_Overlap_5d::LUprecond(Field& v, const Field& w,
                                const int ieo)
{
  //      ieo=0: 1/M_ee, 1: 1/M_oo

  Field t(w);

  int Nin   = w.nin();
  int Nvol2 = w.nvol();
  int Nex   = w.nex();

  Field vx(Nin, Nvol2, 1);
  Field vy(Nin, Nvol2, 1);
  Field vj(Nin, Nvol2, 1);
  Field v_tmp(Nin, Nvol2, 1);

  // --- L^-1 ---
  for (int ip = 0; ip < m_Np; ++ip) {
    int jx = 2 * ip;
    int jy = 2 * ip + 1;

    vx.setpart_ex(0, w, jx);
    vy.setpart_ex(0, w, jy);

    v_tmp = m_fopr_w->mult_gm5(vx);
    vy   += -m_rl[ip] * v_tmp;

    t.setpart_ex(jx, vx, 0);
    t.setpart_ex(jy, vy, 0);
  }

  int j = 2 * m_Np;
  vj = 0.0;
  for (int ip = 0; ip < m_Np; ++ip) {
    int jy = 2 * ip + 1;
    vy.setpart_ex(0, t, jy);
    vj += -m_sl[ip] * vy;
  }
  v_tmp = m_fopr_w->mult_gm5(vj);
  Proj_H_eo(ieo, ieo, vj, v_tmp);
  v_tmp.setpart_ex(0, w, j);
  vj += v_tmp;
  t.setpart_ex(j, vj, 0);

  // --- U^-1 ---
  mult_u0inv(v_tmp, vj, ieo);
  v.setpart_ex(j, v_tmp, 0);

  Proj_H_eo(ieo, ieo, vj, v_tmp);
  t.setpart_ex(j, vj, 0);

  for (int ip = 0; ip < m_Np; ++ip) {
    int jx = 2 * ip;
    int jy = 2 * ip + 1;

    v_tmp.setpart_ex(0, t, jy);
    vj.setpart_ex(0, t, j);
    v_tmp += -m_p_sqrt[ip] * vj;
    vy     = m_fopr_w->mult_gm5(v_tmp);
    vy    *= -1.0 / (1.0 + m_q_sqrt[ip] * m_q_sqrt[ip]);
    v.setpart_ex(jy, vy, 0);

    v_tmp.setpart_ex(0, t, jx);
    v_tmp += m_q_sqrt[ip] * vy;
    vx     = m_fopr_w->mult_gm5(v_tmp);
    v.setpart_ex(jx, vx, 0);
  }
}


//====================================================================
void Fopr_Overlap_5d::Proj_H_eo(const int ieo1, const int ieo2,
                                Field& v, const Field& w)
{
  if (ieo1 == ieo2) {
    v = w;
  } else {
    v = 0.0;
  }

  int    j1, j2;
  double prd_r, prd_i, vr, vi;

  for (int k = 0; k < m_Nsbt; ++k) {
    j1 = k + ieo1 * m_Nsbt;
    j2 = k + ieo2 * m_Nsbt;
    innerprd_c(prd_r, prd_i, m_vk[j2], w);
    add_c(v, m_vk[j1], -prd_r, -prd_i);
  }
}


//====================================================================
void Fopr_Overlap_5d::Proj_L_mult_eo(const int ieo1, const int ieo2,
                                     Field& v, const Field& w)
{
  int    j1, j2;
  double prd_r, prd_i, v_r, v_i;

  v = 0.0;
  for (int k = 0; k < m_Nsbt; ++k) {
    j1 = k + ieo1 * m_Nsbt;
    j2 = k + ieo2 * m_Nsbt;
    innerprd_c(prd_r, prd_i, m_vk[j2], w);
    prd_r *= m_prf[k];
    prd_i *= m_prf[k];
    add_c(v, m_vk[j1], prd_r, prd_i);
  }
}


//====================================================================
void Fopr_Overlap_5d::Calc_Coeff_u0inv()
{
  int Nsbt  = m_Nsbt;
  int Nsbt2 = 2 * m_Nsbt;

  valarray<dcomplex> c(Nsbt2 * Nsbt2);
  valarray<dcomplex> cinv(Nsbt2 * Nsbt2);
  valarray<dcomplex> vprd(Nsbt2 * Nsbt2);
  valarray<dcomplex> W(Nsbt2 * Nsbt2);

  valarray<dcomplex> c_src(Nsbt2);
  valarray<dcomplex> c_x(Nsbt2);

  int   Nin  = m_vk[0].nin();
  int   Nvol = m_vk[0].nvol();
  Field w(Nin, Nvol, 1);
  Field v1(Nin, Nvol, 1);
  Field v2(Nin, Nvol, 1);

  double u0_a    = m_R_parameter + m_p0_parameter + m_u0;
  double u0inv_a = 1.0 / u0_a;

  double a_r, a_i;

  for (int ieo = 0; ieo < 2; ++ieo) {
    //- Determine inner product of eigenvectors
    for (int i = 0; i < Nsbt; ++i) {
      for (int j = 0; j < i + 1; ++j) {
        int i2 = i + ieo * Nsbt;
        int j2 = j + ieo * Nsbt;
        v1 = m_vk[i2];
        v2 = m_vk[j2];
        innerprd_c(a_r, a_i, v1, v2);
        vprd[i + j * Nsbt2] = cmplx(a_r, a_i);
        vprd[i + Nsbt + (j + Nsbt) * Nsbt2] = cmplx(a_r, a_i);
        vprd[j + i * Nsbt2] = cmplx(a_r, -a_i);
        vprd[j + Nsbt + (i + Nsbt) * Nsbt2] = cmplx(a_r, -a_i);
      }
    }

    for (int i = 0; i < Nsbt; ++i) {
      for (int j = 0; j < i + 1; ++j) {
        int i2 = i + ieo * Nsbt;
        int j2 = j + ieo * Nsbt;
        v1 = m_vk[i2];
        w  = m_vk[j2];
        v2 = m_fopr_w->mult_gm5(w);
        innerprd_c(a_r, a_i, v1, v2);
        vprd[i + (j + Nsbt) * Nsbt2] = cmplx(a_r, a_i);
        vprd[i + Nsbt + j * Nsbt2]   = cmplx(a_r, a_i);
        vprd[j + (i + Nsbt) * Nsbt2] = cmplx(a_r, -a_i);
        vprd[j + Nsbt + i * Nsbt2]   = cmplx(a_r, -a_i);
      }
    }

    //- definition of matrix c(i,j)
    c = cmplx(0.0, 0.0);
    for (int i = 0; i < Nsbt2; ++i) {
      c[i + i * Nsbt2] = cmplx(-m_u0, 0.0);
    }

    for (int i = 0; i < Nsbt; ++i) {
      double prf = (m_M0 - 0.5 * m_mq) * (m_ev[i] / fabs(m_ev[i]))
                   - m_p0_parameter * m_ev[i];
      c[i + (i + Nsbt) * Nsbt2] = cmplx(prf, 0.0);
    }

    for (int i = 0; i < Nsbt; ++i) {
      for (int j = Nsbt; j < Nsbt2; ++j) {
        c[i + j * Nsbt2] += cmplx(m_u0, 0.0) * vprd[i + j * Nsbt2];
      }
    }

    //- Definition of matrix W(i,j)
    W = cmplx(0.0, 0.0);
    for (int i = 0; i < Nsbt2; ++i) {
      W[i + i * Nsbt2] = cmplx(u0_a, 0.0);
    }

    for (int i = 0; i < Nsbt2; ++i) {
      for (int j = 0; j < Nsbt2; ++j) {
        for (int k = 0; k < Nsbt2; ++k) {
          W[i + j * Nsbt2] += c[i + k * Nsbt2] * vprd[k + j * Nsbt2];
        }
      }
    }

    //- Solve cinv
    for (int i = 0; i < Nsbt2; ++i) {
      for (int j = 0; j < Nsbt2; ++j) {
        c_src[j] = cmplx(-u0inv_a, 0.0) * c[j + i * Nsbt2];
      }

      Solv_Coeff_u0inv(c_x, W, c_src);

      for (int j = 0; j < Nsbt2; ++j) {
        cinv[j + i * Nsbt2] = c_x[j];
      }
    }

    if (ieo == 0) {
      for (int i = 0; i < Nsbt2 * Nsbt2; ++i) {
        m_u0c_e[i]    = c[i];
        m_u0cinv_e[i] = cinv[i];
      }
    } else {
      for (int i = 0; i < Nsbt2 * Nsbt2; ++i) {
        m_u0c_o[i]    = c[i];
        m_u0cinv_o[i] = cinv[i];
      }
    }
  }
}


//====================================================================
void Fopr_Overlap_5d::innerprd_c(double& prd_r, double& prd_i,
                                 const Field& v, const Field& w)
{
  int size = w.size();

  assert(v.size() == size);

  prd_r = 0.0;
  prd_i = 0.0;

  for (int i = 0; i < size; i += 2) {
    prd_r += v.cmp(i) * w.cmp(i) + v.cmp(i + 1) * w.cmp(i + 1);
    prd_i += v.cmp(i) * w.cmp(i + 1) - v.cmp(i + 1) * w.cmp(i);
  }

  prd_r = Communicator::reduce_sum(prd_r);
  prd_i = Communicator::reduce_sum(prd_i);
}


//====================================================================
void Fopr_Overlap_5d::add_c(Field& v, const Field& w,
                            const double a_r, const double a_i)
{
  int size = w.size();

  assert(v.size() == size);

  double v_r, v_i;
  for (int i = 0; i < size; i += 2) {
    v_r = a_r * w.cmp(i) - a_i * w.cmp(i + 1);
    v_i = a_r * w.cmp(i + 1) + a_i * w.cmp(i);
    v.add(i, v_r);
    v.add(i + 1, v_i);
  }
}


//====================================================================
void Fopr_Overlap_5d::mult_u0inv(Field& v1, const Field& w1,
                                 const int ieo)
{
  if (m_Nsbt == 0) {
    v1  = m_fopr_w->mult_gm5(w1);
    v1 *= 1.0 / (m_R_parameter + m_p0_parameter + m_u0);
  } else {
    int   Nin   = w1.nin();
    int   Nvol2 = w1.nvol();
    Field vt(Nin, Nvol2, 1);

    int                Nsbt  = m_Nsbt;
    int                Nsbt2 = 2 * Nsbt;
    valarray<dcomplex> prd_vb(Nsbt2), coeff(Nsbt2);
    valarray<dcomplex> u0c(Nsbt2 * Nsbt2), u0cinv(Nsbt2 * Nsbt2);

    if (ieo == 0) {
      u0c    = m_u0c_e;
      u0cinv = m_u0cinv_e;
    } else {
      u0c    = m_u0c_o;
      u0cinv = m_u0cinv_o;
    }

    double a_r, a_i;
    for (int k = 0; k < Nsbt; ++k) {
      innerprd_c(a_r, a_i, m_vk[k + ieo * Nsbt], w1);
      prd_vb[k] = cmplx(a_r, a_i);
    }
    vt = m_fopr_w->mult_gm5(w1);
    for (int k = 0; k < Nsbt; ++k) {
      innerprd_c(a_r, a_i, m_vk[k + ieo * Nsbt], vt);
      prd_vb[k + Nsbt] = cmplx(a_r, a_i);
    }

    for (int i = 0; i < Nsbt2; ++i) {
      coeff[i] = cmplx(0.0, 0.0);
      for (int j = 0; j < Nsbt2; ++j) {
        coeff[i] += u0cinv[i + j * Nsbt2] * prd_vb[j];
      }
    }

    double u0inv_a = 1.0 / (m_R_parameter + m_p0_parameter + m_u0);

    vt = u0inv_a * w1;

    for (int k = 0; k < Nsbt; ++k) {
      add_c(vt, m_vk[k + ieo * Nsbt], real(coeff[k]), imag(coeff[k]));
    }

    v1 = m_fopr_w->mult_gm5(vt);

    for (int k = 0; k < Nsbt; ++k) {
      add_c(v1, m_vk[k + ieo * Nsbt],
            real(coeff[k + Nsbt]), imag(coeff[k + Nsbt]));
    }
  }
}


//====================================================================
void Fopr_Overlap_5d::Solv_Coeff_u0inv(valarray<dcomplex>& c_x,
                                       const valarray<dcomplex>& W, const valarray<dcomplex>& c_src)
{
  //- This is an implementation of CG solver.
  int Nsbt  = m_Nsbt;
  int Nsbt2 = 2 * m_Nsbt;

  assert(c_x.size() == Nsbt2);
  assert(c_src.size() == Nsbt2);
  assert(W.size() == (Nsbt2 * Nsbt2));

  valarray<dcomplex> x(Nsbt2);
  valarray<dcomplex> r(Nsbt2);
  valarray<dcomplex> p(Nsbt2);
  valarray<dcomplex> s(Nsbt2);
  valarray<dcomplex> vt(Nsbt2);

  int    Niter = 100;
  double Encg  = 1.e-32;

  double ww    = norm_c(c_src);
  double snorm = 1.0 / ww;

  double rr, rr0;
  int    nconv = -1; // superficial initialization

  s = cmplx(0.0, 0.0);
  for (int i = 0; i < Nsbt2; ++i) {
    for (int j = 0; j < Nsbt2; ++j) {
      s[i] += conj(W[j + i * Nsbt2]) * c_src[j];
    }
  }
  x = s;
  r = s;
  mult_WdagW(s, W, x);

  r -= s;
  p  = r;
  rr = norm_c(r);

  vout.detailed(m_vl, "   init   %16.8e\n", rr * snorm);

  if (rr * snorm < Encg) goto converged;


  for (int iter = 0; iter < Niter; ++iter) {
    mult_WdagW(s, W, p);

    double pap   = innerprd_c(p, s);
    double alpha = rr / pap;

    x  += cmplx(alpha, 0.0) * p;
    r  -= cmplx(alpha, 0.0) * s;
    rr0 = rr;
    rr  = norm_c(r);

    vout.detailed(m_vl, "   %4d   %16.8e\n", iter, rr * snorm);

    if (rr * snorm < Encg) {
      nconv = iter;
      goto converged;
    }

    double beta = rr / rr0;

    p *= cmplx(beta, 0.0);
    p += r;
  }

  nconv = -1;

  vout.crucial(m_vl, "Fopr_Overlap_5d:  NOT CONVERGED.\n");
  abort();


converged:
  vout.detailed(m_vl, "  converged\n");

  s = cmplx(0.0, 0.0);
  for (int i = 0; i < Nsbt2; ++i) {
    for (int j = 0; j < Nsbt2; ++j) {
      s[i] += W[i + j * Nsbt2] * x[j];
    }
  }
  s -= c_src;

  double diff = norm_c(s);
  diff *= snorm;

  c_x = x;

  vout.general(m_vl, "  u0 solver: Nconv = %4d, diff = %12.4e\n", nconv, diff);
}


//====================================================================
void Fopr_Overlap_5d::mult_WdagW(valarray<dcomplex>& v2,
                                 const valarray<dcomplex>& W,
                                 const valarray<dcomplex>& v1)
{
  int size = v1.size();

  assert(v2.size() == size);
  assert(W.size() == (size * size));

  valarray<dcomplex> vt(size);

  vt = cmplx(0.0, 0.0);
  for (int i = 0; i < size; ++i) {
    for (int j = 0; j < size; ++j) {
      vt[i] += W[i + j * size] * v1[j];
    }
  }

  v2 = cmplx(0.0, 0.0);
  for (int i = 0; i < size; ++i) {
    for (int j = 0; j < size; ++j) {
      v2[i] += conj(W[j + i * size]) * vt[j];
    }
  }
}


//====================================================================
double Fopr_Overlap_5d::norm_c(const valarray<dcomplex>& v)
{
  int size = v.size();

  double vv = 0.0;

  for (int i = 0; i < size; ++i) {
    vv += real(v[i]) * real(v[i]) + imag(v[i]) * imag(v[i]);
  }

  return vv;
}


//====================================================================
double Fopr_Overlap_5d::innerprd_c(const valarray<dcomplex>& v,
                                   const valarray<dcomplex>& w)
{
  int size = v.size();

  double vw = 0.0;

  for (int i = 0; i < size; ++i) {
    vw += real(v[i]) * real(w[i]) + imag(v[i]) * imag(w[i]);
  }

  return vw;
}


//====================================================================
//============================================================END=====
