/*!
        @file    $Id:: fopr_Sign.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-07-12 14:43:23 #$

        @version $LastChangedRevision: 929 $
*/

#include "fopr_Sign.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using std::valarray;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_int("number_of_poles", 0);
    param.Register_double("lower_bound", 0.0);
    param.Register_double("upper_bound", 0.0);
    param.Register_int("maximum_number_of_iteration", 0);
    param.Register_double("convergence_criterion_squared", 0.0);

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Fopr.Sign", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Fopr_Sign::Parameters_Fopr_Sign() { append_entry(*this); }
//- end

//====================================================================
void Fopr_Sign::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  int    Np;
  double x_min, x_max;
  int    Niter;
  double Stop_cond;

  int err = 0;
  err += params.fetch_int("number_of_poles", Np);
  err += params.fetch_double("lower_bound", x_min);
  err += params.fetch_double("upper_bound", x_max);
  err += params.fetch_int("maximum_number_of_iteration", Niter);
  err += params.fetch_double("convergence_criterion_squared", Stop_cond);

  if (err) {
    vout.crucial(m_vl, "Fopr_Sign: fetch error, input parameter not found.\n");
    abort();
  }

  set_parameters(Np, x_min, x_max, Niter, Stop_cond);
}


//====================================================================
void Fopr_Sign::set_parameters(int Np, double x_min, double x_max, int Niter,
                               double Stop_cond)
{
  //- print input parameters
  vout.general(m_vl, "Fopr_Sign parameters:\n");
  vout.general(m_vl, "  Np        = %4d\n", Np);
  vout.general(m_vl, "  x_min     = %12.8f\n", x_min);
  vout.general(m_vl, "  x_max     = %12.6f\n", x_max);
  vout.general(m_vl, "  Niter     = %6d\n", Niter);
  vout.general(m_vl, "  Stop_cond = %12.4e\n", Stop_cond);

  //- range check
  int err = 0;
  err += ParameterCheck::non_zero(Np);
  // NB. x_min,x_max == 0 is allowed.
  err += ParameterCheck::non_zero(Niter);
  err += ParameterCheck::square_non_zero(Stop_cond);

  if (err) {
    vout.crucial(m_vl, "Fopr_Sign: parameter range check failed.\n");
    abort();
  }

  //- store values
  m_Np        = Np;
  m_x_min     = x_min;
  m_x_max     = x_max;
  m_Niter     = Niter;
  m_Stop_cond = Stop_cond;

  //- post-process
  m_sigma.resize(m_Np);
  m_cl.resize(2 * m_Np);
  m_bl.resize(m_Np);

  init_parameters();
}


//====================================================================
void Fopr_Sign::set_lowmodes(int Nsbt, valarray<double> *ev,
                             valarray<Field> *vk)
{
  m_Nsbt = Nsbt;
  m_ev   = ev;
  m_vk   = vk;
}


//====================================================================
void Fopr_Sign::tidyup()
{
  delete m_solver;
}


//====================================================================
void Fopr_Sign::init_parameters()
{
  m_sigma.resize(m_Np);
  m_cl.resize(2 * m_Np);
  m_bl.resize(m_Np);

  // Zolotarev coefficient defined
  double              bmax = m_x_max / m_x_min;
  Math_Sign_Zolotarev sign_func(m_Np, bmax);
  sign_func.get_sign_parameters(m_cl, m_bl);

  for (int i = 0; i < m_Np; i++) {
    m_sigma[i] = m_cl[2 * i] * m_x_min * m_x_min;
  }

  for (int i = 0; i < m_Np; i++) {
    vout.general(m_vl, " %3d %12.4e %12.4e %12.4e\n",
                 i, m_cl[i], m_cl[i + m_Np], m_bl[i]);
  }

  int Nin  = m_fopr->field_nin();
  int Nvol = m_fopr->field_nvol();
  int Nex  = m_fopr->field_nex();
  m_xq.resize(m_Np);
  for (int i = 0; i < m_Np; ++i) {
    m_xq[i].reset(Nin, Nvol, Nex);
  }

  m_solver = new Shiftsolver_CG(m_fopr, m_Niter, m_Stop_cond);
}


//====================================================================
void Fopr_Sign::mult(Field& v, const Field& b)
{
  assert(b.nin() == m_fopr->field_nin());
  assert(b.nvol() == m_fopr->field_nvol());
  assert(b.nex() == m_fopr->field_nex());

  assert(v.nin() == m_fopr->field_nin());
  assert(v.nvol() == m_fopr->field_nvol());
  assert(v.nex() == m_fopr->field_nex());

  // vout.general(m_vl, "  Sign function: Nsbt = %d\n",m_Nsbt);

  // Low-mode subtraction
  Field b2(b);
  if (m_Nsbt > 0) subtract_lowmodes(b2);

  // Shiftsolver

  int    Nshift = m_Np;
  int    Nconv;
  double diff;

  //vout.general(m_vl, "    Shiftsolver in sign function\n");
  //vout.general(m_vl, "    Number of shift values = %d\n",m_sigma.size());

  m_fopr->set_mode("DdagD");
  m_solver->solve(m_xq, m_sigma, b2, Nconv, diff);

  //  Field v(b);
  Field w(b);

  v = m_bl[0] * m_xq[0];
  for (int i = 1; i < m_Np; i++) {
    v += m_bl[i] * m_xq[i];
  }

  w = m_fopr->mult(v);
  double coeff = m_cl[2 * m_Np - 1] * m_x_min * m_x_min;
  w += coeff * v;

  m_fopr->set_mode("H");
  v  = m_fopr->mult(w);
  v /= m_x_min;

  if (m_Nsbt > 0) evaluate_lowmodes(v, b);

  //  return v;
}


//====================================================================
void Fopr_Sign::subtract_lowmodes(Field& w)
{
  if ((w.nin() % 2) != 0) abort();

  int    Nin  = w.nin();
  int    Nvol = w.nvol();
  int    Nex  = w.nex();
  double prd_r, prd_i;
  double v_r, v_i;

  for (int k = 0; k < m_Nsbt; ++k) {
    prd_r = 0.0;
    prd_i = 0.0;
    for (int ex = 0; ex < Nex; ++ex) {
      for (int iv = 0; iv < Nvol; ++iv) {
        for (int in = 0; in < Nin; in += 2) {
          prd_r += (*m_vk)[k].cmp(in, iv, ex) * w.cmp(in, iv, ex)
                   + (*m_vk)[k].cmp(in + 1, iv, ex) * w.cmp(in + 1, iv, ex);
          prd_i += (*m_vk)[k].cmp(in, iv, ex) * w.cmp(in + 1, iv, ex)
                   - (*m_vk)[k].cmp(in + 1, iv, ex) * w.cmp(in, iv, ex);
        }
      }
    }

    prd_r = Communicator::reduce_sum(prd_r);
    prd_i = Communicator::reduce_sum(prd_i);

    for (int ex = 0; ex < Nex; ++ex) {
      for (int iv = 0; iv < Nvol; ++iv) {
        for (int in = 0; in < Nin; in += 2) {
          v_r = w.cmp(in, iv, ex) - prd_r * (*m_vk)[k].cmp(in, iv, ex)
                + prd_i * (*m_vk)[k].cmp(in + 1, iv, ex);
          v_i = w.cmp(in + 1, iv, ex) - prd_r * (*m_vk)[k].cmp(in + 1, iv, ex)
                - prd_i * (*m_vk)[k].cmp(in, iv, ex);
          w.set(in, iv, ex, v_r);
          w.set(in + 1, iv, ex, v_i);
        }
      }
    }
  }
}


//====================================================================
void Fopr_Sign::evaluate_lowmodes(Field& x, const Field& w)
{
  if ((w.nin() % 2) != 0) abort();

  int    Nin  = w.nin();
  int    Nvol = w.nvol();
  int    Nex  = w.nex();
  double prd_r, prd_i;
  double v_r, v_i;

  for (int k = 0; k < m_Nsbt; ++k) {
    prd_r = 0.0;
    prd_i = 0.0;
    for (int ex = 0; ex < Nex; ++ex) {
      for (int iv = 0; iv < Nvol; ++iv) {
        for (int in = 0; in < Nin; in += 2) {
          prd_r += (*m_vk)[k].cmp(in, iv, ex) * w.cmp(in, iv, ex)
                   + (*m_vk)[k].cmp(in + 1, iv, ex) * w.cmp(in + 1, iv, ex);
          prd_i += (*m_vk)[k].cmp(in, iv, ex) * w.cmp(in + 1, iv, ex)
                   - (*m_vk)[k].cmp(in + 1, iv, ex) * w.cmp(in, iv, ex);
        }
      }
    }

    prd_r = Communicator::reduce_sum(prd_r);
    prd_i = Communicator::reduce_sum(prd_i);

    double sgn = (*m_ev)[k] / fabs((*m_ev)[k]);
    prd_r *= sgn;
    prd_i *= sgn;

    for (int ex = 0; ex < Nex; ++ex) {
      for (int iv = 0; iv < Nvol; ++iv) {
        for (int in = 0; in < Nin; in += 2) {
          v_r = x.cmp(in, iv, ex) + prd_r * (*m_vk)[k].cmp(in, iv, ex)
                - prd_i * (*m_vk)[k].cmp(in + 1, iv, ex);
          v_i = x.cmp(in + 1, iv, ex) + prd_r * (*m_vk)[k].cmp(in + 1, iv, ex)
                + prd_i * (*m_vk)[k].cmp(in, iv, ex);
          x.set(in, iv, ex, v_r);
          x.set(in + 1, iv, ex, v_i);
        }
      }
    }
  }
}


//====================================================================
double Fopr_Sign::sign_zolotarev(double x)
{
//  cl[2*Np], bl[Np]: coefficients of rational approx.

  double x2R = 0.0;

  for (int l = 0; l < m_Np; l++) {
    x2R += m_bl[l] / (x * x + m_cl[2 * l]);
  }
  x2R = x2R * (x * x + m_cl[2 * m_Np - 1]);

  return x * x2R;
}


//====================================================================
//============================================================END=====
