/*!
        @file    $Id:: fopr_Sign.h #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-07-12 14:43:23 #$

        @version $LastChangedRevision: 929 $
*/

#ifndef FOPR_SIGN_INCLUDED
#define FOPR_SIGN_INCLUDED

#include "fopr.h"
#include "field_G.h"
#include "shiftsolver_CG.h"
#include "math_Sign_Zolotarev.h"

#include "bridgeIO.h"
using Bridge::vout;

//! Sign function of a given fermion operator.

/*!
    This class defines a sign function of a given fermion operator,
    which must be provided at the construction, and thus itself
    being a kind of fermion operator.
    The sign function is approximated with partially fractional
    series whose coefficients are given by Zolotarev approximation.
    When eigenvalues and eigenvectors of the base fermion operator
    are given, contribution of these modes to sign function are
    explicitly calculated and subtracted from the approximation
    formula.
    Number of subtracted modes are initially set to zero, and if
    the function set_lowmodes() is not called, no subtraction is
    performed.
    In the present implementation, shiftsolver is explicitly set
    to CG, which might be generalized.
                                     [20 Dec 2011 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.             [14 Nov 2012 Y.Namekawa]
 */

//- parameters class
class Parameters_Fopr_Sign : virtual public Parameters
{
 public:
  Parameters_Fopr_Sign();
};
//- end

class Fopr_Sign : public Fopr
{
 private:
  int    m_Np;             // number of poles in rational approx.
  double m_x_min, m_x_max; // valid range of approximate sign function
  int    m_Niter;          // max iteration of shiftsolver
  double m_Stop_cond;      // stopping condition of shift solver

  Fopr *m_fopr;

  std::valarray<double> m_cl;
  std::valarray<double> m_bl;
  std::valarray<double> m_sigma;

  //- low-mode subtraction
  int m_Nsbt;
  std::valarray<double> *m_ev;
  std::valarray<Field>  *m_vk;

  //- shiftsolver
  Shiftsolver_CG       *m_solver;
  std::valarray<Field> m_xq;

 public:

  Fopr_Sign(Fopr *fopr)
    : Fopr(), m_fopr(fopr)
  {
    m_Nsbt = 0;
    m_ev   = 0;
    m_vk   = 0;
  }

  ~Fopr_Sign()
  {
    tidyup();
  }

  void set_parameters(const Parameters& params);
  void set_parameters(int Np, double x_min, double x_max,
                      int Niter, double Stop_cond);

  void set_config(Field *U)
  {
    m_fopr->set_config(U);
  }

  void set_mode(std::string mode)
  {
    m_fopr->set_mode(mode);
  }

  std::string get_mode() const
  {
    return m_fopr->get_mode();
  }

  void tidyup();

  void set_lowmodes(int Nsbt, std::valarray<double> *,
                    std::valarray<Field> *);

  void mult(Field& v, const Field& f);

  const Field mult(const Field& f)
  {
    Field v(f.nin(), f.nvol(), f.nex());

    mult(v, f);

    return v;
  }

  void mult_dag(Field& v, const Field& f)
  {
    mult(v, f);
  }

  const Field mult_dag(const Field& f)
  {
    Field v(f.nin(), f.nvol(), f.nex());

    mult_dag(v, f);

    return v;
  }

  int field_nvol() { return m_fopr->field_nvol(); }
  int field_nin() { return m_fopr->field_nin(); }
  int field_nex() { return m_fopr->field_nex(); }

 private:
  void init_parameters();

  void subtract_lowmodes(Field&);
  void evaluate_lowmodes(Field&, const Field&);

  double sign_zolotarev(double x);
};
#endif
