/*!
        @file    $Id:: force_F_Overlap_Nf2.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-07-12 14:43:23 #$

        @version $LastChangedRevision: 929 $
*/

#include "force_F_Overlap_Nf2.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using std::valarray;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("quark_mass", 0.0);
    param.Register_double("domain_wall_height", 0.0);
    param.Register_int("number_of_poles", 0);
    param.Register_double("lower_bound", 0.0);
    param.Register_double("upper_bound", 0.0);
    param.Register_int("maximum_number_of_iteration", 0);
    param.Register_double("convergence_criterion_squared", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Force.F_Overlap_Nf2", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Force_F_Overlap_Nf2::Parameters_Force_F_Overlap_Nf2() { append_entry(*this); }
//- end

//====================================================================
void Force_F_Overlap_Nf2::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        mq, M0;
  int           Np;
  double        x_min, x_max;
  int           Niter_ms;
  double        Stop_cond_ms;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("quark_mass", mq);
  err += params.fetch_double("domain_wall_height", M0);
  err += params.fetch_int("number_of_poles", Np);
  err += params.fetch_double("lower_bound", x_min);
  err += params.fetch_double("upper_bound", x_max);
  err += params.fetch_int("maximum_number_of_iteration", Niter_ms);
  err += params.fetch_double("convergence_criterion_squared", Stop_cond_ms);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "Force_F_Overlap_Nf2: fetch error, input parameter not found.\n");
    abort();
  }


  set_parameters(mq, M0, Np, x_min, x_max, Niter_ms, Stop_cond_ms, bc);
}


//====================================================================
void Force_F_Overlap_Nf2::set_parameters(const double M0, const double mq,
                                         const int Np, const double x_min, const double x_max,
                                         const int Niter_ms, const double Stop_cond_ms,
                                         const valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Force_F_Overlap_Nf2 paramters:\n");
  vout.general(m_vl, "  M0           = %10.6f\n", M0);
  vout.general(m_vl, "  mq           = %10.6f\n", mq);
  vout.general(m_vl, "  Np           = %4d\n", Np);
  vout.general(m_vl, "  x_min        = %12.8f\n", x_min);
  vout.general(m_vl, "  x_max        = %12.6f\n", x_max);
  vout.general(m_vl, "  Niter_ms     = %6d\n", Niter_ms);
  vout.general(m_vl, "  Stop_cond_ms = %12.4e\n", Stop_cond_ms);
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", mu, bc[mu]);
  }

  //- range check
  int err = 0;
  err += ParameterCheck::non_zero(M0);
  err += ParameterCheck::non_zero(mq);
  err += ParameterCheck::non_zero(Np);
  // NB. x_min,x_max == 0 is allowed.
  err += ParameterCheck::non_zero(Niter_ms);
  err += ParameterCheck::square_non_zero(Stop_cond_ms);

  if (err) {
    vout.crucial(m_vl, "Force_F_Overlap_Nf2: parameter range check failed.\n");
    abort();
  }

  assert(bc.size() == Ndim);

  //- store values
  m_M0           = M0;
  m_mq           = mq;
  m_Np           = Np;
  m_x_min        = x_min;
  m_x_max        = x_max;
  m_Niter_ms     = Niter_ms;
  m_Stop_cond_ms = Stop_cond_ms;

  m_boundary.resize(Ndim);
  for (int mu = 0; mu < Ndim; ++mu) {
    m_boundary[mu] = bc[mu];
  }

  //- post-process
  m_kappa = 0.5 / (4.0 - m_M0);

  m_fopr_w = new Fopr_Wilson();
  m_fopr_w->set_parameters(m_kappa, m_boundary);
  m_fopr_w->set_config(m_U);

  m_force_w = new Force_F_Wilson_Nf2;
  m_force_w->set_parameters(m_kappa, m_boundary);
  m_force_w->set_config(m_U);

  //- Zolotarev coefficients
  m_sigma.resize(m_Np);
  m_cl.resize(2 * m_Np);
  m_bl.resize(m_Np);
}


//====================================================================
Field Force_F_Overlap_Nf2::force_core(const Field& psi)
{
  int Nc       = CommonParameters::Nc();
  int Nd       = CommonParameters::Nd();
  int Nvol     = CommonParameters::Nvol();
  int Lvol     = CommonParameters::Lvol();
  int Ndim     = CommonParameters::Ndim();
  int size_psf = Lvol * Nc * Nd;

  //- solving psi = (H^2)^(-1) phi
  Field_F psi5(Nvol, 1);

  psi5 = (Field_F)m_fopr_w->mult_gm5(psi);

  double ff = m_M0 * m_M0 - 0.25 * m_mq * m_mq;
  psi5 *= ff;

  // force determination

  int   Nin = Nc * Nc * 2;
  Field force(Nin, Nvol, Ndim);

  force  = force_core1(psi, psi5);
  force *= -1.0;

  return force;
}


//====================================================================
Field Force_F_Overlap_Nf2::force_core1(const Field_F& psi,
                                       const Field_F& psi5)
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();
  int NinG = 2 * Nc * Nc;
  int NinF = 2 * Nc * Nd;

  //- Zolotarev coefficient defined
  double              bmax = m_x_max / m_x_min;
  Math_Sign_Zolotarev sign_func(m_Np, bmax);

  sign_func.get_sign_parameters(m_cl, m_bl);

  for (int i = 0; i < m_Np; ++i) {
    m_sigma[i] = m_cl[2 * i] * m_x_min * m_x_min;
  }

  for (int i = 0; i < m_Np; ++i) {
    vout.general(m_vl, " %3d %12.4e %12.4e %12.4e\n",
                 i, m_cl[i], m_cl[i + m_Np], m_bl[i]);
  }

  //- Shiftsolver
  int Nshift = m_Np;

  valarray<Field> psi_l(Nshift);
  for (int i = 0; i < Nshift; ++i) {
    psi_l[i].reset(NinF, Nvol, 1);
  }

  valarray<Field> psi5_l(Nshift);
  for (int i = 0; i < Nshift; ++i) {
    psi5_l[i].reset(NinF, Nvol, 1);
  }

  int    Nconv;
  double diff;

  vout.general(m_vl, "Shift solver in MD\n");
  vout.general(m_vl, "  Number of shift values = %d\n", m_sigma.size());

  m_fopr_w->set_mode("DdagD");

  Shiftsolver_CG *solver = new Shiftsolver_CG(m_fopr_w,
                                              m_Niter_ms, m_Stop_cond_ms);

  solver->solve(psi_l, m_sigma, psi, Nconv, diff);
  solver->solve(psi5_l, m_sigma, psi5, Nconv, diff);

  delete solver;

  Field force(NinG, Nvol, Ndim);
  force = 0.0;

  Field_F psid(Nvol, 1);
  Field_F psi5d(Nvol, 1);
  psid  = 0.0;
  psi5d = 0.0;

  Field   force1(NinG, Nvol, Ndim);
  Field_F w1(Nvol, 1);
  Field_F w2(Nvol, 1);
  Field_F w3(Nvol, 1);
  Field_F w4(Nvol, 1);

  for (int ip = 0; ip < m_Np; ++ip) {
    psid.addpart_ex(0, psi_l[ip], 0, m_bl[ip]);
    psi5d.addpart_ex(0, psi5_l[ip], 0, m_bl[ip]);

    // w1 = m_fopr_w->H(psi_l[ip]);
    // w2 = m_fopr_w->H(psi5_l[ip]);
    m_fopr_w->set_mode("H");
    w1 = (Field_F)m_fopr_w->mult(psi_l[ip]);
    w2 = (Field_F)m_fopr_w->mult(psi5_l[ip]);

    double coeff_l = (m_cl[2 * ip] - m_cl[2 * m_Np - 1])
                     * m_bl[ip] * m_x_min;

    //- first term
    // w3 = m_fopr_w->H(w2);
    w3 = (Field_F)m_fopr_w->mult(w2);
    //    force1 = m_force_w->force_core1(w3,(Field_F)psi_l[ip]);
    w4 = (Field_F)psi_l[ip];

    force1 = m_force_w->force_core1(w3, w4);
    force += coeff_l * force1;
    force1 = m_force_w->force_core1(w2, w1);
    force += coeff_l * force1;

    //- second term
    // w3 = m_fopr_w->H(w1);
    w3 = (Field_F)m_fopr_w->mult(w1);
    //    force1 = m_force_w->force_core1(w3,(Field_F)psi5_l[ip]);
    w4 = (Field_F)psi5_l[ip];

    force1 = m_force_w->force_core1(w3, w4);
    force += coeff_l * force1;
    force1 = m_force_w->force_core1(w1, w2);
    force += coeff_l * force1;
  }

  double coeff1 = m_cl[2 * m_Np - 1] * m_x_min * m_x_min;
  double coeff2 = 1.0 / m_x_min;

  //- first term
  // w1 = m_fopr_w->H(psid);
  // w2 = m_fopr_w->H(w1);
  w1 = (Field_F)m_fopr_w->mult(psid);
  w2 = (Field_F)m_fopr_w->mult(w1);
  w2.addpart_ex(0, psid, 0, coeff1);

  force1 = m_force_w->force_core1(psi5, w2);
  force += coeff2 * force1;

  //- second term
  // w1 = m_fopr_w->H(psi5d);
  // w2 = m_fopr_w->H(w1);
  w1 = (Field_F)m_fopr_w->mult(psi5d);
  w2 = (Field_F)m_fopr_w->mult(w1);
  w2.addpart_ex(0, psi5d, 0, coeff1);

  force1 = m_force_w->force_core1(psi, w2);
  force += coeff2 * force1;


  return force;
}


//====================================================================
//===========================================================END======
