/*!
        @file    $Id:: force_F_Staggered_eo.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-07-12 14:43:23 #$

        @version $LastChangedRevision: 929 $
*/

#include "force_F_Staggered_eo.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using std::valarray;
using Bridge::vout;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("quark_mass", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Force.F_Staggered_eo", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Force_F_Staggered_eo::Parameters_Force_F_Staggered_eo() { append_entry(*this); }
//- end

//====================================================================
void Force_F_Staggered_eo::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        mq;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("quark_mass", mq);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "Force_F_Staggered_eo: fetch error, input parameter not found.\n");
    abort();
  }


  set_parameters(mq, bc);
}


//====================================================================
void Force_F_Staggered_eo::set_parameters(double mq, const valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Parameters of staggered fermion force:\n");
  vout.general(m_vl, "  mq   = %8.4f\n", mq);
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", mu, bc[mu]);
  }

  //- range check
  int err = 0;
  err += ParameterCheck::non_zero(mq);

  if (err) {
    vout.crucial(m_vl, "Force_Staggered_eo: parameter range check failed.\n");
    abort();
  }

  assert(bc.size() == Ndim);

  //- store values
  m_mq = mq;

  // m_boundary.resize(Ndim);  // already resized in the constructor.
  for (int mu = 0; mu < Ndim; ++mu) {
    m_boundary[mu] = bc[mu];
  }

  //- post-process
  m_fopr_ks->set_parameters(m_mq, m_boundary);
}


//====================================================================
Field Force_F_Staggered_eo::force_core(const Field& eta)
{
  int Nc    = CommonParameters::Nc();
  int Nvol  = CommonParameters::Nvol();
  int Ndim  = CommonParameters::Ndim();
  int Nvol2 = Nvol / 2;

  assert(eta.nin() == 2 * Nc);
  assert(eta.nvol() == Nvol2);
  assert(eta.nex() == 1);

  Field_G  force(Nvol, Ndim);
  Field_G  force_eo(Nvol, Ndim);
  Field_G  force_e(Nvol2, Ndim), force_o(Nvol2, Ndim);
  Mat_SU_N ut(Nc);

  Field_F_1spinor zeta(Nvol2, 1);
  zeta = (Field_F_1spinor)m_fopr_ks->Meo(eta, 1);

  force_e  = (Field_G)force_udiv1((Field_F_1spinor)eta, zeta, 0);
  force_o  = (Field_G)force_udiv1(zeta, (Field_F_1spinor)eta, 1);
  force_o *= -1.0;

  int gx;
  for (int mu = 0; mu < Ndim; ++mu) {
    for (int site = 0; site < Nvol2; ++site) {
      gx = m_index_eo.site(site, 0);
      ut = m_Ueo->mat(gx, mu) * force_e.mat(site, mu);
      ut.at();
      ut *= -2.0;
      force_eo.set_mat(gx, mu, ut);

      gx = m_index_eo.site(site, 1);
      ut = m_Ueo->mat(gx, mu) * force_o.mat(site, mu);
      ut.at();
      ut *= -2.0;
      force_eo.set_mat(gx, mu, ut);
    }
  }

  m_index_eo.reverseField(force, force_eo);

  return (Field)force;
}


//====================================================================
Field Force_F_Staggered_eo::force_udiv(const Field& eta)
{
  int Nc    = CommonParameters::Nc();
  int Nvol  = CommonParameters::Nvol();
  int Ndim  = CommonParameters::Ndim();
  int Nvol2 = Nvol / 2;

  assert(eta.nin() == 2 * Nc);
  assert(eta.nvol() == Nvol2);
  assert(eta.nex() == 1);

  Field_G force(Nvol, Ndim);
  Field_G force_e(Nvol2, Ndim), force_o(Nvol2, Ndim);

  Field_F_1spinor zeta(Nvol2, 1);
  zeta = (Field_F_1spinor)m_fopr_ks->Meo(eta, 1);

  force_e  = (Field_G)force_udiv1((Field_F_1spinor)eta, zeta, 0);
  force_o  = (Field_G)force_udiv1(zeta, (Field_F_1spinor)eta, 1);
  force_o *= -1.0;

  m_index_eo.reverseField(force, force_e, 0);
  m_index_eo.reverseField(force, force_o, 1);

  return (Field)force;
}


//====================================================================
Field Force_F_Staggered_eo::force_udiv1(const Field_F_1spinor& zeta,
                                        const Field_F_1spinor& eta, int ieo)
{
  int Nc    = CommonParameters::Nc();
  int Nvol  = CommonParameters::Nvol();
  int Ndim  = CommonParameters::Ndim();
  int Nvol2 = Nvol / 2;

  Field_G forceh(Nvol2, Ndim);

  Field_F_1spinor eta2(Nvol2, 1);
  Mat_SU_N        ut(Nc);
  double          utr, uti;

  for (int mu = 0; mu < Ndim; ++mu) {
    m_shift_eo.backward_h(eta2, eta, m_boundary[mu], mu, ieo);
    m_fopr_ks->mult_staggered_phase(eta2, mu, ieo);

    for (int site = 0; site < Nvol2; ++site) {
      for (int c1 = 0; c1 < Nc; ++c1) {
        for (int c2 = 0; c2 < Nc; ++c2) {
          utr = zeta.cmp_r(c2, site) * eta2.cmp_r(c1, site)
                + zeta.cmp_i(c2, site) * eta2.cmp_i(c1, site);
          uti = zeta.cmp_r(c2, site) * eta2.cmp_i(c1, site)
                - zeta.cmp_i(c2, site) * eta2.cmp_r(c1, site);
          utr *= -0.5 / m_mq;
          uti *= -0.5 / m_mq;
          ut.set(c1, c2, utr, uti);
        }
      }

      forceh.set_mat(site, mu, ut);
    }
  }

  return (Field)forceh;
}


//====================================================================
//============================================================END=====
