/*!
        @file    $Id:: force_F_TMWilson_Nf2.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-07-12 14:43:23 #$

        @version $LastChangedRevision: 929 $
*/

#include "force_F_TMWilson_Nf2.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using std::valarray;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("hopping_parameter", 0.0);
    param.Register_double("twisted_mass", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Force.F_TMWilson_Nf2", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Force_F_TMWilson_Nf2::Parameters_Force_F_TMWilson_Nf2() { append_entry(*this); }
//- end

//====================================================================
void Force_F_TMWilson_Nf2::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        kappa, tw_mass;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("hopping_parameter", kappa);
  err += params.fetch_double("twisted_mass", tw_mass);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "Force_F_TMWilson: fetch error, input parameter not found.\n");
    abort();
  }


  set_parameters(kappa, tw_mass, bc);
}


//====================================================================
void Force_F_TMWilson_Nf2::set_parameters(const double kappa, const double tw_mass,
                                          const valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Parameters of twisted mass fermion operator:\n");
  vout.general(m_vl, "  kappa   = %8.4f\n", kappa);
  vout.general(m_vl, "  tw_mass = %8.4f\n", tw_mass);
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", mu, bc[mu]);
  }

  //- range check
  // NB. kappa,tw_mass == 0 is allowed.
  assert(bc.size() == Ndim);

  //- store values
  m_kappa   = kappa;
  m_tw_mass = tw_mass;

  m_boundary.resize(Ndim);
  for (int mu = 0; mu < Ndim; ++mu) {
    m_boundary[mu] = bc[mu];
  }

  //- post-process
  m_fopr_tw->set_parameters(m_kappa, m_tw_mass, m_boundary);
}


//====================================================================
Field Force_F_TMWilson_Nf2::force_core(const Field& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();
  int NinG = 2 * Nc * Nc;

  Field_G  force(Nvol, Ndim);
  Field_G  force1(Nvol, Ndim);
  Mat_SU_N ut(Nc);


  force1 = force_udiv(eta);

  for (int mu = 0; mu < Ndim; ++mu) {
    force.mult_Field_Gnn(mu, *m_U, mu, force1, mu);
    force.at_Field_G(mu);
  }
  force *= -2.0;

  return (Field)force;
}


//====================================================================
Field Force_F_TMWilson_Nf2::force_core1(const Field& zeta,
                                        const Field& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();
  int NinG = 2 * Nc * Nc;

  Field_G  force(Nvol, Ndim);
  Field_G  force1(Nvol, Ndim);
  Mat_SU_N ut(Nc);

  force1 = force_udiv1(zeta, eta);

  for (int mu = 0; mu < Ndim; ++mu) {
    force.mult_Field_Gnn(mu, *m_U, mu, force1, mu);
    force.at_Field_G(mu);
  }
  force *= -2.0;

  return (Field)force;
}


//====================================================================
Field Force_F_TMWilson_Nf2::force_udiv(const Field& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G force(Nvol, Ndim);
  Field_G force1(Nvol, Ndim);

  Field_F zeta(Nvol, 1);

  m_fopr_tw->set_mode("H");
  zeta = (Field_F)m_fopr_tw->mult(eta);

  force  = (Field_G)force_udiv1(zeta, (Field_F)eta);
  force1 = (Field_G)force_udiv1((Field_F)eta, zeta);
  force += force1;

  return (Field)force;
}


//====================================================================
Field Force_F_TMWilson_Nf2::force_udiv1(const Field& zeta,
                                        const Field& eta)
{
  return force_udiv1((Field_F)zeta, (Field_F)eta);
}


//====================================================================
Field Force_F_TMWilson_Nf2::force_udiv1(const Field_F& zeta,
                                        const Field_F& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G force(Nvol, Ndim);
  Field_G force1(Nvol, 1);

  Field_F  eta2(Nvol, 1), eta3(Nvol, 1);
  Mat_SU_N ut(Nc), ut2(Nc);
  double   utr, uti;

  for (int mu = 0; mu < Ndim; ++mu) {
    eta2 = m_fopr_tw->mult_gm5p(mu, eta);
    eta3.mult_Field_Gd(0, *m_U, mu, eta2, 0);
    tensorProd_Field_F(force1, zeta, eta3);
    force.setpart_ex(mu, force1, 0);
  }

  force *= -m_kappa;

  return (Field)force;
}


//====================================================================
//============================================================END=====
