/*!
        @file    $Id:: force_F_Wilson_Nf2_Isochemical.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-07-12 14:43:23 #$

        @version $LastChangedRevision: 929 $
*/

#include "force_F_Wilson_Nf2_Isochemical.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using std::valarray;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("hopping_parameter", 0.0);
    param.Register_double("isospin_chemical_potential", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Force.F_Wilson_Nf2_Isochemical", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Force_F_Wilson_Nf2_Isochemical::Parameters_Force_F_Wilson_Nf2_Isochemical() { append_entry(*this); }
//- end

//====================================================================
void Force_F_Wilson_Nf2_Isochemical::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        kappa, mu;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("hopping_parameter", kappa);
  err += params.fetch_double("isospin_chemical_potential", mu);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "Force_F_Wilson_Isochemical: fetch error, input parameter not found.\n");
    abort();
  }


  set_parameters(kappa, mu, bc);
}


//====================================================================
void Force_F_Wilson_Nf2_Isochemical::set_parameters(const double kappa, const double mu,
                                                    const valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Parameters of Wilson fermion operator:\n");
  vout.general(m_vl, "  kappa = %8.4f\n", kappa);
  vout.general(m_vl, "  mu    = %8.4f\n", mu);
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", mu, bc[mu]);
  }

  //- range check
  // NB. kappa,mu == 0 is allowed.
  assert(bc.size() == Ndim);

  //- store values
  m_kappa  = kappa;
  m_mu     = mu;
  m_exp_mu = exp(mu);

  m_boundary.resize(Ndim);
  for (int dir = 0; dir < Ndim; ++dir) {
    m_boundary[dir] = bc[dir];
  }

  //- post-process
  m_fopr_w->set_parameters(m_kappa, m_mu, m_boundary);
}


//====================================================================
Field Force_F_Wilson_Nf2_Isochemical::force_core(const Field& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();
  int NinG = 2 * Nc * Nc;

  Field_G force(Nvol, Ndim);
  Field_G force1(Nvol, Ndim);

  force1 = force_udiv(eta);

  for (int mu = 0; mu < Ndim; ++mu) {
    force.mult_Field_Gnn(mu, *m_U, mu, force1, mu);
    force.at_Field_G(mu);
  }
  force *= -2.0;

  return (Field)force;
}


//====================================================================
Field Force_F_Wilson_Nf2_Isochemical::force_core1(const Field& zeta,
                                                  const Field& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();
  int NinG = 2 * Nc * Nc;

  Field_G force(Nvol, Ndim);
  Field_G force1(Nvol, Ndim);

  force1 = force_udiv1(zeta, eta);

  for (int mu = 0; mu < Ndim; ++mu) {
    force.mult_Field_Gnn(mu, *m_U, mu, force1, mu);
    force.at_Field_G(mu);
  }
  force *= -2.0;

  return (Field)force;
}


//====================================================================
Field Force_F_Wilson_Nf2_Isochemical::force_udiv(const Field& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();
  int NinG = 2 * Nc * Nc;

  Field force(NinG, Nvol, Ndim);
  Field force2(NinG, Nvol, Ndim);

  Field_F zeta(Nvol, 1);

  m_fopr_w->set_mode("H");
  zeta = (Field_F)m_fopr_w->mult(eta);

  set_mode("H");
  force = force_udiv1(zeta, (Field_F)eta);

  set_mode("Hdag");
  force2 = force_udiv1((Field_F)eta, zeta);

  force += force2;

  return (Field)force;
}


//====================================================================
Field Force_F_Wilson_Nf2_Isochemical::force_udiv1(const Field& zeta,
                                                  const Field& eta)
{
  return force_udiv1((Field_F)zeta, (Field_F)eta);
}


//====================================================================
Field Force_F_Wilson_Nf2_Isochemical::force_udiv1(const Field_F& zeta,
                                                  const Field_F& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G force(Nvol, Ndim);
  Field_G force1(Nvol, 1);

  Field_F eta2(Nvol, 1), eta3(Nvol, 1);

  for (int dir = 0; dir < Ndim - 1; ++dir) {
    eta2 = m_fopr_w->mult_gm5p(dir, eta);
    eta3.mult_Field_Gd(0, *m_U, dir, eta2, 0);
    eta3 *= -m_kappa;
    tensorProd_Field_F(force1, zeta, eta3);
    force.setpart_ex(dir, force1, 0);
  }

  {
    int dir = Ndim - 1;
    eta2 = m_fopr_w->mult_gm5p(dir, eta);
    eta3.mult_Field_Gd(0, *m_U, dir, eta2, 0);
    if (m_mode == "H") {
      eta3 *= -(m_kappa * m_exp_mu);
    } else if (m_mode == "Hdag") {
      eta3 *= -(m_kappa / m_exp_mu);
    } else {
      vout.crucial(m_vl, "Force_F_Wilson_Nf2_Isochemical: illegal mode.\n");
      abort();
    }
    tensorProd_Field_F(force1, zeta, eta3);
    force.setpart_ex(dir, force1, 0);
  }

  return (Field)force;
}


//====================================================================
//============================================================END=====
