/*!
        @file    $Id:: fprop_Overlap_5d.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-04-27 12:28:50 #$

        @version $LastChangedRevision: 875 $
*/

#include "fprop_Overlap_5d.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

//- parameter entry
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("quark_mass", 0.0);
    param.Register_double("domain_wall_height", 0.0);
    param.Register_int("number_of_poles", 0);
    param.Register_double("lower_bound", 0.0);
    param.Register_double("upper_bound", 0.0);
    param.Register_int("maximum_number_of_iteration", 0);
    param.Register_double("convergence_criterion_squared", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Fprop_Overlap_5d", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Fprop_Overlap_5d::Parameters_Fprop_Overlap_5d() { append_entry(*this); }
//- end

//====================================================================
void Fprop_Overlap_5d::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        mq, M0;
  int           Np;
  double        x_min, x_max;
  int           Niter_ms;
  double        Stop_cond_ms;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("quark_mass", mq);
  err += params.fetch_double("domain_wall_height", M0);
  err += params.fetch_int("number_of_poles", Np);
  err += params.fetch_double("lower_bound", x_min);
  err += params.fetch_double("upper_bound", x_max);
  err += params.fetch_int("maximum_number_of_iteration", Niter_ms);
  err += params.fetch_double("convergence_criterion_squared", Stop_cond_ms);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "Fprop_Overlap_5d: fetch error, input parameter not found.\n");
    abort();
  }


  set_parameters(mq, M0, Np, x_min, x_max, Niter_ms, Stop_cond_ms, bc);
}


//====================================================================
void Fprop_Overlap_5d::set_parameters(const Parameters& params_overlap,
                                      const Parameters& params_solver)
{
  double             mq           = params_overlap.get_double("quark_mass");
  double             M0           = params_overlap.get_double("domain_wall_height");
  int                Np           = params_overlap.get_int("number_of_poles");
  double             x_min        = params_overlap.get_double("lower_bound");
  double             x_max        = params_overlap.get_double("upper_bound");
  int                Niter_ms     = params_overlap.get_int("maximum_number_of_iteration");
  double             Stop_cond_ms = params_overlap.get_double("convergence_criterion_squared");
  std::valarray<int> bc           = params_overlap.get_int_vector("boundary_condition");

  set_parameters(mq, M0, Np, x_min, x_max, Niter_ms, Stop_cond_ms, bc);
  m_params_solver = params_solver;
}


//====================================================================
void Fprop_Overlap_5d::set_parameters(const double mq, const double M0,
                                      const int Np, const double x_min, const double x_max,
                                      const int Niter_ms, const double Stop_cond_ms,
                                      const valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Fprop_Overlap_5d paramters:\n");
  vout.general(m_vl, "  mq           = %10.6f\n", mq);
  vout.general(m_vl, "  M0           = %10.6f\n", M0);
  vout.general(m_vl, "  Np           = %4d\n", Np);
  vout.general(m_vl, "  x_min        = %12.8f\n", x_min);
  vout.general(m_vl, "  x_max        = %12.6f\n", x_max);
  vout.general(m_vl, "  Niter_ms     = %6d\n", Niter_ms);
  vout.general(m_vl, "  Stop_cond_ms = %12.4e\n", Stop_cond_ms);
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", mu, bc[mu]);
  }

  //- range check
  int err = 0;
  err += ParameterCheck::non_zero(mq);
  err += ParameterCheck::non_zero(M0);
  err += ParameterCheck::non_zero(Np);
  // NB. x_min,x_max == 0 is allowed.
  err += ParameterCheck::non_zero(Niter_ms);
  err += ParameterCheck::square_non_zero(Stop_cond_ms);

  if (err) {
    vout.crucial(m_vl, "Fprop_Overlap_5d: parameter range check failed.\n");
    abort();
  }

  assert(bc.size() == Ndim);

  //- store values
  m_mq           = mq;
  m_M0           = M0;
  m_Np           = Np;
  m_x_min        = x_min;
  m_x_max        = x_max;
  m_Niter_ms     = Niter_ms;
  m_Stop_cond_ms = Stop_cond_ms;

  m_boundary.resize(Ndim);
  assert(bc.size() == Ndim);
  for (int mu = 0; mu < Ndim; ++mu) {
    m_boundary[mu] = bc[mu];
  }

  m_Nsbt = 0;
}


//====================================================================
void Fprop_Overlap_5d::set_env()
{
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  //- setting even-odd envoronement
  m_Ueo = new Field_G(Nvol, Ndim);
  m_index_eo.convertField(*m_Ueo, *m_conf);

  //- setting Wilson fermion operator
  double kappa = 0.5 / (4.0 - m_M0);
  m_fopr_w_eo = new Fopr_Wilson_eo("Dirac");
  m_fopr_w_eo->set_parameters(kappa, m_boundary);
  // NB. U is converted to Ueo in fopr_eo->set_config
  // m_fopr_w_eo->set_config(m_Ueo);
  m_fopr_w_eo->set_config(m_conf);

  //- setting 5D overlap fermion operator
  m_fopr_ov5d = new Fopr_Overlap_5d(m_fopr_w_eo);
  m_fopr_ov5d->set_parameters(m_mq, m_M0, m_Np, m_x_min, m_x_max,
                              m_Niter_ms, m_Stop_cond_ms, m_boundary);
  m_fopr_ov5d->set_lowmodes(m_Nsbt, m_ev, m_vk);
}


//====================================================================
void Fprop_Overlap_5d::calc_fprop(Field& xq, const Field& b)
{
  set_env();

  double snorm = 1.0 / b.norm2();

  Field  v1((Field)b);
  int    nconv, nconv1, nconv2;
  double diff, diff1, diff2;

  v1 = m_fopr_w_eo->mult_gm5(b);
  solve_overlap_5D_1(xq, v1, nconv, diff, snorm);
  //v1 = m_fopr_ov->D(xq);
  m_fopr_ov->set_mode("D");
  m_fopr_ov->D(v1, xq);

  v1 -= b;
  double vv = v1 * v1;
  vv = vv * snorm;

  vout.general(m_vl, "5D overlap solver total:\n");
  vout.general(m_vl, "  Nconv_eo(ov) = %10d\n", nconv);
  vout.general(m_vl, "  Diff_eo(ov)  = %16.8e\n", diff);
  vout.general(m_vl, "  Diff(ov)     = %16.8e\n", vv);

  delete m_Ueo;
  delete m_fopr_w_eo;
  delete m_fopr_ov5d;
}


//====================================================================
void Fprop_Overlap_5d::calc_H2inv(Field& xq, const Field& b)
{
  set_env();

  double snorm = 1.0 / b.norm2();

  Field  v1((Field)b);
  int    nconv, nconv1, nconv2;
  double diff, diff1, diff2;

  solve_overlap_5D_1(v1, (Field)b, nconv1, diff1, snorm);
  solve_overlap_5D_1(xq, v1, nconv2, diff2, snorm);

  nconv = nconv1 + nconv2;
  diff  = diff1 + diff2;

  //  v1 = m_fopr_ov->DdagD(xq);
  m_fopr_ov->set_mode("DdagD");
  m_fopr_ov->mult(v1, xq);

  v1 -= b;
  double vv = v1 * v1;
  vv = vv * snorm;

  vout.general(m_vl, "  5D overlap solver for Q_ov total:\n");
  vout.general(m_vl, "  5D overlap solver total:\n");
  vout.general(m_vl, "  Nconv_eo(ov) = %10d\n", nconv);
  vout.general(m_vl, "  Diff_eo(ov)  = %16.8e\n", diff);
  vout.general(m_vl, "  Diff(ov)     = %16.8e\n", vv);

  delete m_Ueo;
  delete m_fopr_w_eo;
  delete m_fopr_ov5d;
}


//====================================================================
void Fprop_Overlap_5d::solve_overlap_5D_1(Field& xq, const Field& b,
                                          int& nconv, double& diff, double& snorm)
{
  // ####  parameter setup  ####
  int Npl = 2 * m_Np + 1;

  int Nin   = b.nin();
  int Nvol  = b.nvol();
  int Nvol2 = Nvol / 2;

  const string str_solver_type = m_params_solver.get_string("solver_type");

  // ####  object setup  ####
  Solver *solver = Solver::New(str_solver_type, m_fopr_ov5d);

  solver->set_parameters(m_params_solver);

  Field v(Nin, Nvol2, Npl);
  Field s(Nin, Nvol2, Npl);
  Field t(Nin, Nvol2, Npl);
  Field p(Nin, Nvol2, Npl);
  Field x(Nin, Nvol2, Npl);

  Field be(Nin, Nvol2, 1);
  Field bo(Nin, Nvol2, 1);

  m_index_eo.convertField(be, b, 0);
  m_index_eo.convertField(bo, b, 1);

  v = 0.0;
  v.setpart_ex(2 * m_Np, be, 0);
  m_fopr_ov5d->LUprecond(t, v, 0);

  v = 0.0;
  v.setpart_ex(2 * m_Np, bo, 0);
  m_fopr_ov5d->LUprecond(p, v, 1);
  m_fopr_ov5d->Mopr_5d_eo(v, p, 0);
  m_fopr_ov5d->LUprecond(p, v, 0);

  t -= p;

  s = m_fopr_ov5d->DD_5d_eo(t, -1);


  solver->solve(x, s, nconv, diff);


  //- Check
  v = 0.0;
  v.setpart_ex(2 * m_Np, be, 0);
  m_fopr_ov5d->LUprecond(s, v, 0);

  v = 0.0;
  v.setpart_ex(2 * m_Np, bo, 0);
  m_fopr_ov5d->LUprecond(p, v, 1);
  m_fopr_ov5d->Mopr_5d_eo(t, p, 0);
  m_fopr_ov5d->LUprecond(p, t, 0);

  s -= p;

  p     = m_fopr_ov5d->DD_5d_eo(x, 1);
  p    -= s;
  diff  = p * p;
  diff *= snorm;

  //- Solution
  Field xqe(Nin, Nvol2, 1);
  Field xqo(Nin, Nvol2, 1);
  Field xt(Nin, Nvol2, 1);

  xqe.setpart_ex(0, x, 2 * m_Np);

  v = 0.0;
  v.setpart_ex(2 * m_Np, bo, 0);
  m_fopr_ov5d->LUprecond(p, v, 1);
  xqo.setpart_ex(0, p, 2 * m_Np);

  m_fopr_ov5d->Mopr_5d_eo(p, x, 1);
  m_fopr_ov5d->LUprecond(t, p, 1);
  xt.setpart_ex(0, t, 2 * m_Np);
  xqo -= xt;

  m_index_eo.reverseField(xq, xqe, 0);
  m_index_eo.reverseField(xq, xqo, 1);

  vout.general(m_vl, "  solver_eo1(ov):\n");
  vout.general(m_vl, "  Nconv_eo1(ov) = %10d\n", nconv);
  vout.general(m_vl, "  Diff_eo1(ov)  = %16.8e\n", diff);


  // ####  tidy up  ####
  delete solver;
}


//====================================================================
//============================================================END=====
