/*!
        @file    $Id: test_HMC_Domainwall_Leapfrog_Nf2_DWPV.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate: 2013-04-27 12:28:50 #$

        @version $LastChangedRevision: 875 $
*/

#include "parameterManager_YAML.h"
#include "parameters_factory.h"

#include "bridgeIO.h"
using Bridge::vout;

#include "gaugeConfig.h"

#include "action_G_Plaq.h"
#include "action_G_Rectangle.h"

#include "fopr_Domainwall.h"
#include "force_F_Domainwall.h"

#include "randomNumbers_Mseries.h"

#include "director_Smear.h"
#include "fopr_Smeared.h"
#include "force_F_Smeared.h"

#include "action_F_Standard_lex.h"
#include "fprop_Standard_lex.h"
#include "action_F_Ratio_lex.h"

#include "forceSmear.h"
#include "projection.h"
#include "smear.h"
#include "solver.h"

#include "hmc_General.h"
#include "builder_Integrator.h"
#include "integrator.h"

#ifdef USE_TEST
#include "test.h"
#endif

#ifdef USE_TESTMANAGER_AUTOREGISTER
#include "testManager.h"
#endif

//====================================================================
//! Test of HMC update for domain-wall fermion operator.

/*!
    This class tests HMC update for dynamical domain-wall fermions.
                                         [12 Apr 2012 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.                 [14 Nov 2012 Y.Namekawa]
    Selectors are implemented.           [03 Mar 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
 */

namespace Test_HMC_Domainwall {
  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_HMC_Domainwall_Leapfrog_Nf2_DWPV.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_HMC_Domainwall : public Parameters {
     public:
      Parameters_Test_HMC_Domainwall()
      {
        Register_string("gauge_config_status", "NULL");

        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("gauge_config_type_output", "NULL");
        Register_string("config_filename_output", "NULL");

        Register_int("trajectory_number", 0);
        Register_int("trajectory_number_step", 0);
        Register_int("save_config_interval", 0);

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int leapfrog_Nf2_DWPV(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
    static const bool is_registered = TestManager::RegisterTest(
      "HMC.Domainwall.Leapfrog_Nf2_DWPV",
      leapfrog_Nf2_DWPV
      );
  }
#endif

  //====================================================================
  int leapfrog_Nf2_DWPV(void)
  {
    // #####  parameter setup  #####
    int Nc   = CommonParameters::Nc();
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();

    Parameters *params_test       = new Parameters_Test_HMC_Domainwall;
    Parameters *params_action_G   = ParametersFactory::New("Action.G_Rectangle");
    Parameters *params_proj       = ParametersFactory::New("Projection");
    Parameters *params_smear      = ParametersFactory::New("Smear");
    Parameters *params_dr_smear   = ParametersFactory::New("Director_Smear");
    Parameters *params_dwall      = ParametersFactory::New("Fopr.Domainwall");
    Parameters *params_dwall_pv   = ParametersFactory::New("Fopr.Domainwall");
    Parameters *params_solver_MD  = ParametersFactory::New("Solver");
    Parameters *params_solver_H   = ParametersFactory::New("Solver");
    Parameters *params_integrator = ParametersFactory::New("Builder_Integrator");
    Parameters *params_hmc        = ParametersFactory::New("HMC.General");

    Parameters *params_all = new Parameters;

    params_all->Register_Parameters("Test_HMC_Domainwall", params_test);
    params_all->Register_Parameters("Action_G_Rectangle", params_action_G);
    params_all->Register_Parameters("Fopr_Domainwall", params_dwall);
    params_all->Register_Parameters("Fopr_Domainwall_PauliVillars", params_dwall_pv);
    params_all->Register_Parameters("Projection", params_proj);
    params_all->Register_Parameters("Smear", params_smear);
    params_all->Register_Parameters("Director_Smear", params_dr_smear);
    params_all->Register_Parameters("Solver_MD", params_solver_MD);
    params_all->Register_Parameters("Solver_H", params_solver_H);
    params_all->Register_Parameters("Builder_Integrator", params_integrator);
    params_all->Register_Parameters("HMC_General", params_hmc);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    const string str_gconf_write  = params_test->get_string("gauge_config_type_output");
    const string writefile        = params_test->get_string("config_filename_output");
    int          iconf            = params_test->get_int("trajectory_number");
    int          Ntraj            = params_test->get_int("trajectory_number_step");
    const int    i_save_conf      = params_test->get_int("save_config_interval");
    const string str_vlevel       = params_test->get_string("verbose_level");
#ifdef USE_TEST
    const double expected_result = params_test->get_double("expected_result");
#endif

    const string str_gmset_type     = params_dwall->get_string("gamma_matrix_type");
    const string str_proj_type      = params_proj->get_string("projection_type");
    const string str_smear_type     = params_smear->get_string("smear_type");
    const string str_solver_MD_type = params_solver_MD->get_string("solver_type");
    const string str_solver_H_type  = params_solver_H->get_string("solver_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status   = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read     = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile       = %s\n", readfile.c_str());
    vout.general(vl, "  gconf_write    = %s\n", str_gconf_write.c_str());
    vout.general(vl, "  writefile      = %s\n", writefile.c_str());
    vout.general(vl, "  iconf          = %d\n", iconf);
    vout.general(vl, "  Ntraj          = %d\n", Ntraj);
    vout.general(vl, "  i_save_conf    = %d\n", i_save_conf);
    vout.general(vl, "  vlevel         = %s\n", str_vlevel.c_str());

    vout.general(vl, "  gmset_type     = %s\n", str_gmset_type.c_str());
    vout.general(vl, "  proj_type      = %s\n", str_proj_type.c_str());
    vout.general(vl, "  smear_type     = %s\n", str_smear_type.c_str());
    vout.general(vl, "  solver_MD_type = %s\n", str_solver_MD_type.c_str());
    vout.general(vl, "  solver_H_type  = %s\n", str_solver_H_type.c_str());
    vout.general(vl, "\n");

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_read);
    err += ParameterCheck::non_NULL(readfile);
    err += ParameterCheck::non_NULL(str_gconf_write);
    err += ParameterCheck::non_NULL(writefile);
    err += ParameterCheck::non_zero(iconf);
    err += ParameterCheck::non_zero(Ntraj);
    err += ParameterCheck::non_zero(i_save_conf);

    if (err) {
      vout.crucial(vl, "Test_HMC_Domainwall: Input parameters have not been set.\n");
      abort();
    }


    // #####  object setup  #####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Start_cold") {
      gconf_read->set_cold((Field *)U);
    } else {
      vout.crucial(vl, "Test_HMC_Domainwall: unsupported gconf status \"%s\".\n", str_gconf_status.c_str());
      abort();
    }


    Action_G_Rectangle *action_G = new Action_G_Rectangle;
    action_G->set_parameters(*params_action_G);


    Projection *proj  = Projection::New(str_proj_type);
    Smear      *smear = Smear::New(str_smear_type, proj);
    smear->set_parameters(*params_smear);

    ForceSmear *force_smear = ForceSmear::New(str_smear_type, proj);
    force_smear->set_parameters(*params_smear);

    Director_Smear *dr_smear = new Director_Smear((Smear *)smear);
    dr_smear->set_parameters(*params_dr_smear);


    Fopr_Wilson     *fopr_w  = new Fopr_Wilson(str_gmset_type);
    Fopr_Domainwall *fopr_dw = new Fopr_Domainwall(fopr_w);
    fopr_dw->set_parameters(*params_dwall);

    Force_F_Domainwall *force_fopr_dw = new Force_F_Domainwall();
    force_fopr_dw->set_parameters(*params_dwall);

    Fopr_Smeared    *fopr_smear       = new Fopr_Smeared((Fopr *)fopr_dw, dr_smear);
    Force_F_Smeared *force_fopr_smear = new Force_F_Smeared(
      (Force *)force_fopr_dw, (ForceSmear *)force_smear, dr_smear);


    Solver *solver_MD = Solver::New(str_solver_MD_type, fopr_smear);
    solver_MD->set_parameters(*params_solver_MD);
    Fprop *fprop_MD = new Fprop_Standard_lex(solver_MD);

    Solver *solver_H = Solver::New(str_solver_H_type, fopr_smear);
    solver_H->set_parameters(*params_solver_H);
    Fprop *fprop_H = new Fprop_Standard_lex(solver_H);


    //- Pauli-Villars
    Fopr_Domainwall *fopr_dw_pv = new Fopr_Domainwall(fopr_w);
    fopr_dw_pv->set_parameters(*params_dwall_pv);

    Force_F_Domainwall *force_fopr_dw_pv = new Force_F_Domainwall();
    force_fopr_dw_pv->set_parameters(*params_dwall_pv);

    Fopr_Smeared    *fopr_smear_pv       = new Fopr_Smeared((Fopr *)fopr_dw_pv, dr_smear);
    Force_F_Smeared *force_fopr_smear_pv =
      new Force_F_Smeared((Force *)force_fopr_dw_pv,
                          (ForceSmear *)force_smear, dr_smear);


    Solver *solver_H_pv = Solver::New(str_solver_H_type, fopr_smear_pv);
    solver_H_pv->set_parameters(*params_solver_H);
    Fprop *fprop_H_pv = new Fprop_Standard_lex(solver_H_pv);


    Action_F_Ratio_lex *action_F
      = new Action_F_Ratio_lex((Fopr *)fopr_smear_pv, (Force *)force_fopr_smear_pv,
                               (Fopr *)fopr_smear, (Force *)force_fopr_smear,
                               fprop_H_pv,
                               fprop_MD, fprop_H);


    valarray<Action *> actions(2);
    actions[0] = (Action *)action_F;
    actions[1] = (Action *)action_G;

    valarray<Director *> directors(1);
    directors[0] = (Director *)dr_smear;

    RandomNumbers *rand = new RandomNumbers_Mseries(iconf);

    Builder_Integrator *builder
      = new Builder_Integrator(actions, directors);
    builder->set_parameters(*params_integrator);
    Integrator *integrator = builder->build();

    HMC_General hmc(actions, directors, integrator, rand);
    hmc.set_parameters(*params_hmc);


    // ####  Execution main part  ####
    vout.general(vl, "HMC: Ntraj = %d\n", Ntraj);

    double result = 0.0;
    for (int traj = 0; traj < Ntraj; ++traj) {
      vout.general(vl, "\n");
      vout.general(vl, "---------------------------------------------------\n");
      vout.general(vl, "traj = %d\n", traj);

      result = hmc.update(*U);
    }


    // #####  tidy up  #####
    delete params_test;
    delete params_action_G;
    delete params_dwall;
    delete params_dwall_pv;
    delete params_proj;
    delete params_smear;
    delete params_dr_smear;
    delete params_solver_MD;
    delete params_solver_H;
    delete params_hmc;
    delete params_all;

    delete U;
    delete gconf_read;

    delete rand;

    delete action_G;
    delete action_F;
    delete builder; // added by s.motoki[12/06/05].

    delete force_fopr_dw_pv;
    delete       fopr_dw_pv;
    delete force_fopr_dw;
    delete       fopr_dw;
    delete       fopr_w;

    delete proj;
    delete            smear;
    delete      force_smear;
    delete         dr_smear;
    delete       fopr_smear;
    delete force_fopr_smear;
    delete       fopr_smear_pv;
    delete force_fopr_smear_pv;

    delete solver_MD;
    delete  fprop_MD;
    delete solver_H;
    delete  fprop_H;
    delete solver_H_pv;
    delete  fprop_H_pv;


#ifdef USE_TEST
    return Test::verify(expected_result, result);

#else
    return EXIT_SUCCESS;
#endif
  }
} // namespace Test_HMC_Domainwall
