/*!
        @file    $Id: test_QuarkNumberSusceptibility_Clover_Isochemical.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate: 2013-01-22 13:51:53 #$

        @version $LastChangedRevision: 889 $
*/

#include "parameterManager_YAML.h"
#include "parameters_factory.h"

#include "bridgeIO.h"
using Bridge::vout;

#include "gaugeConfig.h"

#include "fopr_Clover_Isochemical.h"
#include "fprop_Standard_lex.h"
#include "noiseVector_Z2.h"
#include "quarkNumberSusceptibility_Wilson.h"

#include "randomNumbers_Mseries.h"

#include "director_Smear.h"
#include "fopr_Smeared.h"

#include "projection.h"
#include "smear.h"
#include "solver.h"

#ifdef USE_TEST
#include "test.h"
#endif

#ifdef USE_TESTMANAGER_AUTOREGISTER
#include "testManager.h"
#endif

//====================================================================
//! Quark number susceptibility for the Wilson-type fermion.

/*!
    This class tests measurements of the traces which is used to determine
    the quark number susceptibility.
                                [02 Sep 2012 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.        [14 Nov 2012 Y.Namekawa]
    Selectors are implemented.  [03 Mar 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
 */

namespace Test_QuarkNumSuscept {
  // test-private parameters
  namespace {
    const std::string filename_input  = "test_QuarkNumberSusceptibility_Clover_Isochemical.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_QuarkNumSuscept_Clover_Isochemical : public Parameters {
     public:
      Parameters_Test_QuarkNumSuscept_Clover_Isochemical()
      {
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_int("seed_for_noise", 0);
        Register_int("number_of_noises", 0);

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  // prototype declaration
  int quark_num_suscept(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
    static const bool is_registered = TestManager::RegisterTest(
      "QuarkNumberSusceptibility.Clover_Isochemical",
      quark_num_suscept
      );
  }
#endif

  //====================================================================
  int quark_num_suscept(void)
  {
    // #####  parameter setup  #####
    int Nc   = CommonParameters::Nc();
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();
    int NinG = 2 * Nc * Nc;

    Parameters *params_test     = new Parameters_Test_QuarkNumSuscept_Clover_Isochemical;
    Parameters *params_clover   = ParametersFactory::New("Fopr.Clover_Isochemical");
    Parameters *params_proj     = ParametersFactory::New("Projection");
    Parameters *params_smear    = ParametersFactory::New("Smear");
    Parameters *params_dr_smear = ParametersFactory::New("Director_Smear");
    Parameters *params_solver   = ParametersFactory::New("Solver");

    Parameters *params_all = new Parameters;

    params_all->Register_Parameters("Test_QuarkNumSuscept_Clover_Isochemical", params_test);
    params_all->Register_Parameters("Fopr_Clover_Isochemical", params_clover);
    params_all->Register_Parameters("Projection", params_proj);
    params_all->Register_Parameters("Smear", params_smear);
    params_all->Register_Parameters("Director_Smear", params_dr_smear);
    params_all->Register_Parameters("Solver", params_solver);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_read = params_test->get_string("gauge_config_type_input");
    const string readfile       = params_test->get_string("config_filename_input");
    int          i_seed_noise   = params_test->get_int("seed_for_noise");
    const string str_vlevel     = params_test->get_string("verbose_level");
#ifdef USE_TEST
    const double expected_result = params_test->get_double("expected_result");
#endif

    const string str_gmset_type  = params_clover->get_string("gamma_matrix_type");
    const string str_proj_type   = params_proj->get_string("projection_type");
    const string str_smear_type  = params_smear->get_string("smear_type");
    const string str_solver_type = params_solver->get_string("solver_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_read   = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile     = %s\n", readfile.c_str());
    vout.general(vl, "  i_seed_noise = %d\n", i_seed_noise);
    vout.general(vl, "  vlevel       = %s\n", str_vlevel.c_str());
    vout.general(vl, "  gmset_type   = %s\n", str_gmset_type.c_str());
    vout.general(vl, "  proj_type    = %s\n", str_proj_type.c_str());
    vout.general(vl, "  smear_type   = %s\n", str_smear_type.c_str());
    vout.general(vl, "  solver_type  = %s\n", str_solver_type.c_str());

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_read);
    err += ParameterCheck::non_NULL(readfile);
    err += ParameterCheck::non_zero(i_seed_noise);

    if (err) {
      vout.crucial(vl, "Test_QuarkNumSuscept: Input parameters have not been set.\n");
      abort();
    }


    // #####  object setup  #####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);
    gconf_read->read_file((Field *)U, readfile);
    // gconf->set_cold((Field*)U);

    Fopr_Clover_Isochemical *fopr_c = new Fopr_Clover_Isochemical(str_gmset_type);
    fopr_c->set_parameters(*params_clover);

    Projection *proj  = Projection::New(str_proj_type);
    Smear      *smear = Smear::New(str_smear_type, proj);
    smear->set_parameters(*params_smear);
    Director_Smear *dr_smear = new Director_Smear((Smear *)smear);
    dr_smear->set_parameters(*params_dr_smear);

    Fopr_Smeared *fopr_smear = new Fopr_Smeared((Fopr *)fopr_c, dr_smear);
    fopr_smear->set_config(U);


    //- Random number is initialized with a parameter specified by iseed
    RandomNumbers *rand = new RandomNumbers_Mseries(i_seed_noise);
    NoiseVector   *nv   = new NoiseVector_Z2(rand);

    Solver *solver = Solver::New(str_solver_type, fopr_smear);
    solver->set_parameters(*params_solver);
    Fprop *fprop_lex = new Fprop_Standard_lex(solver);

    QuarkNumberSusceptibility_Wilson *quark_suscept
      = new QuarkNumberSusceptibility_Wilson(fopr_smear, fprop_lex, nv);
    quark_suscept->set_parameters(*params_test);


    // ####  Execution main part  ####
    double result = 0.0;
    result = quark_suscept->measure();


    // #####  tidy up  #####
    delete U;
    delete gconf_read;

    delete rand;
    delete nv;
    delete quark_suscept;

    delete fopr_c;
    delete fprop_lex;

    delete proj;
    delete      smear;
    delete   dr_smear;
    delete fopr_smear;

    delete solver;

    delete params_test;
    delete params_clover;
    delete params_proj;
    delete params_smear;
    delete params_dr_smear;
    delete params_all;


#ifdef USE_TEST
    return Test::verify(expected_result, result);

#else
    return EXIT_SUCCESS;
#endif
  }
} // namespace Test_QuarkNumSuscept
