/*!
        @file    $Id: test_Spectrum_CRSMatrix_Domainwall.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate: 2013-01-22 22:08:29 #$

        @version $LastChangedRevision: 935 $
*/

#include "parameterManager_YAML.h"

#include "bridgeIO.h"
using Bridge::vout;

#include "fopr_CRS.h"
#include "fopr_Wilson.h"
#include "fopr_Domainwall.h"

#include "gaugeConfig.h"

#include "source_4spinor_Local.h"

#include "solver_CG.h"

#ifdef USE_TEST
#include "test.h"
#endif

#ifdef USE_TESTMANAGER_AUTOREGISTER
#include "testManager.h"
#endif

//====================================================================
//! Test of CRS matrix format.

/*!
    This class generates and tests the CRS matrix format data.
    Matrix data in CRS format are generated for the following
    fermion operators together with sample source and solution
    vectors.

    The egenrated data can be tested using the function
    CRSsolver() whether really solved result satisfy the linear
    equation.
                                [24 Dec 2011 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.        [14 Nov 2012 Y.Namekawa]
 */

namespace Test_Spectrum_CRSMatrix {
  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_Spectrum_CRSMatrix_Domainwall.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_Spectrum_CRSMatrix : public Parameters {
     public:
      Parameters_Test_Spectrum_CRSMatrix()
      {
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("matrix_output", "NULL");
        Register_string("source_output", "NULL");
        Register_string("solution_output", "NULL");

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int domainwall(void);

  int CRSsolver(
    const string& solution,
    const string& matrix,
    const string& source,
    double& result       /* return value */
    );


#ifdef USE_TESTMANAGER_AUTOREGISTER
#ifdef USE_MPI
  // this test runs only in single-node environment.
#else
  //- NB. CRS test is skipped, because it is time-consuming.
  // namespace {
  //   static const bool is_registered = TestManager::RegisterTest(
  //     "Spectrum.CRSMatrix.Domainwall",
  //     domainwall
  //     );
  // }
#endif
#endif

  //====================================================================
  int domainwall(void)
  {
    // ####  parameter setup  ####
    int Nc   = CommonParameters::Nc();
    int Nd   = CommonParameters::Nd();
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();

    Parameters_Test_Spectrum_CRSMatrix params_test;

    Parameters params_all;

    params_all.Register_Parameters("Test_Spectrum_CRSMatrix", &params_test);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, &params_all);

    const string str_gconf_read = params_test.get_string("gauge_config_type_input");
    const string readfile       = params_test.get_string("config_filename_input");
    const string matrix_file    = params_test.get_string("matrix_output");
    const string source_file    = params_test.get_string("source_output");
    const string solution_file  = params_test.get_string("solution_output");
    const string str_vlevel     = params_test.get_string("verbose_level");
#ifdef USE_TEST
    const double expected_result = params_test.get_double("expected_result");
#endif

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_read      = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile        = %s\n", readfile.c_str());
    vout.general(vl, "  matrix_output   = %s\n", matrix_file.c_str());
    vout.general(vl, "  source_output   = %s\n", source_file.c_str());
    vout.general(vl, "  solution_output = %s\n", solution_file.c_str());
    vout.general(vl, "  vlevel          = %s\n", str_vlevel.c_str());
    vout.general(vl, "\n");

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_read);
    err += ParameterCheck::non_NULL(readfile);
    err += ParameterCheck::non_NULL(matrix_file);
    err += ParameterCheck::non_NULL(source_file);
    err += ParameterCheck::non_NULL(solution_file);

    if (err) {
      vout.crucial(vl, "Test_Spectrum_CRSMatrix: Input parameters have not been set.\n");
      abort();
    }


    // ####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);
    gconf_read->read_file((Field *)U, readfile);
    // gconf_read->set_cold((Field*)U);


    // ####  object setup  #####
    Fopr_Wilson     *fopr_w  = new Fopr_Wilson;
    Fopr_Domainwall *fopr_dw = new Fopr_Domainwall(fopr_w);
    double          mq       = 0.1;
    double          M0       = 1.6;
    int             Ns       = 8;
    valarray<int>   boundary(Ndim);
    boundary[0] = 1;
    boundary[1] = 1;
    boundary[2] = 1;
    boundary[3] = -1;
    fopr_dw->set_parameters(mq, M0, Ns, boundary);
    fopr_dw->set_config(U);

    fopr_dw->set_mode("D");
    Fopr_CRS *fopr_crs = new Fopr_CRS(fopr_dw);
    fopr_crs->write_matrix(matrix_file);

    valarray<int> source_position(Ndim);
    source_position[0] = 0;
    source_position[1] = 0;
    source_position[2] = 0;
    source_position[3] = 0;
    Source_4spinor_Local *source = new Source_4spinor_Local();
    source->set_parameters(source_position);

    //- CGNE solver
    int    Niter     = 2000;
    double Stop_cond = 1.0e-28;
    Solver *solver   = new Solver_CG(fopr_dw);
    solver->set_parameters(Niter, Stop_cond);


    // ####  Execution main part  ####
    Field_F b(Nvol, 1);
    Field_F xq(Nvol, 1);

    int Nin = 2 * Nc * Nd;
    int Nex = Ns;

    Field bt(Nin, Nvol, 1);
    Field xt(Nin, Nvol, 1), xt2(Nin, Nvol, 1);

    Field b_5d(Nin, Nvol, Nex), bt_5d(Nin, Nvol, Nex);
    Field xq_5d(Nin, Nvol, Nex);

    double diff = 0.0L;
    {
      int ispin = 0;
      {
        int icolor = 0;
        {
          int ch = -1;
          // for(int ispin = 0; ispin < Nd; ++ispin){
          //  for(int icolor = 0; icolor < Nc; ++icolor){
          //   for(int ch = -1; ch <= 1; ch = ch+2){

          //     source->mksrc(b,ispin,icolor);
          source->set(b, icolor, ispin);

          bt = fopr_dw->mult_chproj_4d(b, ch);

          bt_5d = 0.0;
          if (ch == -1) {
            bt_5d.setpart_ex(Nex - 1, bt, 0);
          } else {
            bt_5d.setpart_ex(0, bt, 0);
          }

          //- save the source vector
          bt_5d.write_text(source_file);

          fopr_dw->set_mode("Ddag");
          b_5d = fopr_dw->mult(bt_5d);

          vout.general(vl, " b_5d norm2 = %14.6e\n", b_5d.norm2());

          int Nconv;
          fopr_dw->set_mode("DdagD");
          solver->solve(xq_5d, b_5d, Nconv, diff);
          vout.general(vl, "  chirality: %2d\n", ch);
          vout.general(vl, "    Nconv = %d,  diff = %16.8e\n", Nconv, diff);

          //- save the solution vector
          xq_5d.write_text(solution_file);
        }
      }
    }


    double result;
    CRSsolver(solution_file, matrix_file, source_file, result);


    // ####  tidy up  ####
    delete U;
    delete gconf_read;

    delete solver;
    delete source;
    delete fopr_crs;
    delete fopr_dw;
    delete fopr_w;


#ifdef USE_TEST
    return Test::verify(expected_result, result);

#else
    return EXIT_SUCCESS;
#endif
  }
} // namespace Test_Spectrum_CRSMatrix
