/*!
        @file    $Id: test_Spectrum_CRSMatrix_Overlap5d.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate: 2013-01-22 22:08:29 #$

        @version $LastChangedRevision: 935 $
*/

#include "parameterManager_YAML.h"

#include "bridgeIO.h"
using Bridge::vout;

// #include "index_lex.h"
#include "index_eo.h"

#include "fopr_CRS.h"
#include "fopr_Wilson_eo.h"
#include "fopr_Overlap_5d.h"
#include "fopr_Wilson.h"
#include "fopr_Overlap.h"

#include "gaugeConfig.h"

#include "source_4spinor_Local.h"

#include "solver_CG.h"

#ifdef USE_TEST
#include "test.h"
#endif

#ifdef USE_TESTMANAGER_AUTOREGISTER
#include "testManager.h"
#endif

//====================================================================
//! Test of CRS matrix format.

/*!
    This class generates and tests the CRS matrix format data.
    Matrix data in CRS format are generated for the following
    fermion operators together with sample source and solution
    vectors.

    The egenrated data can be tested using the function
    CRSsolver() whether really solved result satisfy the linear
    equation.
                                [24 Dec 2011 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.        [14 Nov 2012 Y.Namekawa]
 */

namespace Test_Spectrum_CRSMatrix {
  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_Spectrum_CRSMatrix_Overlap5d.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_Spectrum_CRSMatrix : public Parameters {
     public:
      Parameters_Test_Spectrum_CRSMatrix()
      {
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("matrix_output", "NULL");
        Register_string("source_output", "NULL");
        Register_string("solution_output", "NULL");

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int overlap_5d(void);

  int CRSsolver(
    const string& solution,
    const string& matrix,
    const string& source,
    double& result       /* return value */
    );


#ifdef USE_TESTMANAGER_AUTOREGISTER
#ifdef USE_MPI
  // this test runs only in single-node environment.
#else
  //- NB. CRS test is skipped, because it is time-consuming.
  // namespace {
  //   static const bool is_registered = TestManager::RegisterTest(
  //     "Spectrum.CRSMatrix.Overlap5d",
  //     overlap_5d
  //     );
  // }
#endif
#endif

  //====================================================================
  int overlap_5d(void)
  {
    // ####  parameter setup  ####
    int Nc   = CommonParameters::Nc();
    int Nd   = CommonParameters::Nd();
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();

    Parameters_Test_Spectrum_CRSMatrix params_test;

    Parameters params_all;

    params_all.Register_Parameters("Test_Spectrum_CRSMatrix", &params_test);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, &params_all);

    const string str_gconf_read = params_test.get_string("gauge_config_type_input");
    const string readfile       = params_test.get_string("config_filename_input");
    const string matrix_file    = params_test.get_string("matrix_output");
    const string source_file    = params_test.get_string("source_output");
    const string solution_file  = params_test.get_string("solution_output");
    const string str_vlevel     = params_test.get_string("verbose_level");
#ifdef USE_TEST
    const double expected_result = params_test.get_double("expected_result");
#endif

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_read      = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile        = %s\n", readfile.c_str());
    vout.general(vl, "  matrix_output   = %s\n", matrix_file.c_str());
    vout.general(vl, "  source_output   = %s\n", source_file.c_str());
    vout.general(vl, "  solution_output = %s\n", solution_file.c_str());
    vout.general(vl, "  vlevel          = %s\n", str_vlevel.c_str());

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_read);
    err += ParameterCheck::non_NULL(readfile);
    err += ParameterCheck::non_NULL(matrix_file);
    err += ParameterCheck::non_NULL(source_file);
    err += ParameterCheck::non_NULL(solution_file);

    if (err) {
      vout.crucial(vl, "Test_Spectrum_CRSMatrix: Input parameters have not been set.\n");
      abort();
    }


    // ####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);
    gconf_read->read_file((Field *)U, readfile);
    // gconf->set_cold((Field*)U);

    Field_G  *Ueo = new Field_G(Nvol, Ndim);
    Index_eo index_eo;
    index_eo.convertField(*Ueo, *U);


    // ####  object setup  #####
    //- overlap parameters
    double        mq           = 0.2;
    double        M0           = 1.6;
    int           Np           = 4;
    double        x_min        = 0.10;
    double        x_max        = 8.0;
    double        Niter_ms     = 1000;
    double        Stop_cond_ms = 1.0e-28;
    valarray<int> boundary(Ndim);
    boundary[0] = 1;
    boundary[1] = 1;
    boundary[2] = 1;
    boundary[3] = -1;


    Fopr_Wilson_eo *fopr_w_eo = new Fopr_Wilson_eo("Dirac");
    double         kappa      = 0.5 / (4.0 - M0);
    fopr_w_eo->set_parameters(kappa, boundary);
    fopr_w_eo->set_config(Ueo);

    Fopr_Overlap_5d *fopr_ov5d = new Fopr_Overlap_5d(fopr_w_eo);
    fopr_ov5d->set_parameters(mq, M0, Np, x_min, x_max,
                              Niter_ms, Stop_cond_ms, boundary);
    fopr_ov5d->set_mode("D_eo");

    Fopr_CRS *fopr_crs = new Fopr_CRS(fopr_ov5d);
    fopr_crs->write_matrix(matrix_file);


    //- NB. fopr_ov is used only for check
    Fopr_Wilson *fopr_w = new Fopr_Wilson;
    fopr_w->set_parameters(kappa, boundary);
    fopr_w->set_config(U);

    Fopr_Overlap *fopr_ov = new Fopr_Overlap(fopr_w);
    fopr_ov->set_parameters(mq, M0, Np, x_min, x_max,
                            Niter_ms, Stop_cond_ms, boundary);


    int    Niter_5d     = 1000;
    double Stop_cond_5d = 1.0e-28;
    Solver *solver      = new Solver_CG(fopr_ov5d);
    solver->set_parameters(Niter_5d, Stop_cond_5d);

    valarray<int> source_position(Ndim);
    source_position[0] = 0;
    source_position[1] = 0;
    source_position[2] = 0;
    source_position[3] = 0;
    Source_4spinor_Local *source = new Source_4spinor_Local();
    source->set_parameters(source_position);


    // ####  Execution main part  ####
    valarray<Field_F> sq(Nc * Nd);
    for (int cd = 0; cd < Nc * Nd; ++cd) {
      sq[cd] = 0.0;
    }

    Field_F b;
    Field   xq((Field)b);
    Field   v1(xq);

    int Nin   = fopr_ov5d->field_nin();
    int Nvol2 = fopr_ov5d->field_nvol();
    int Npl   = fopr_ov5d->field_nex();

    Field v(Nin, Nvol2, Npl);
    Field t(Nin, Nvol2, Npl);
    Field p(Nin, Nvol2, Npl);
    Field s(Nin, Nvol2, Npl);
    Field x(Nin, Nvol2, Npl);

    Field be(Nin, Nvol2, 1);
    Field bo(Nin, Nvol2, 1);
    Field xqe(Nin, Nvol2, 1);
    Field xqo(Nin, Nvol2, 1);
    Field xt(Nin, Nvol2, 1);

    int    Nconv;
    double diff;

    double vv = 0.0L;

    // { int ispin = 0;
    // { int icolor = 0;
    for (int ispin = 0; ispin < Nd; ++ispin) {
      for (int icolor = 0; icolor < Nc; ++icolor) {
        source->set(b, icolor, ispin);

        double snorm = 1.0 / b.norm2();

        index_eo.convertField(be, b, 0);
        index_eo.convertField(bo, b, 1);

        v = 0.0;
        v.setpart_ex(2 * Np, be, 0);
        fopr_ov5d->LUprecond(t, v, 0);

        v = 0.0;
        v.setpart_ex(2 * Np, bo, 0);
        fopr_ov5d->LUprecond(p, v, 1);
        fopr_ov5d->Mopr_5d_eo(v, p, 0);
        fopr_ov5d->LUprecond(p, v, 0);

        t -= p;
        t.write_text(source_file);
        s = fopr_ov5d->DD_5d_eo(t, -1);


        fopr_ov5d->set_mode("DdagD_eo");
        solver->solve(x, s, Nconv, diff);

        x.write_text(solution_file);

        //- Solution
        xqe.setpart_ex(0, x, 2 * Np);

        v = 0.0;
        v.setpart_ex(2 * Np, bo, 0);
        fopr_ov5d->LUprecond(p, v, 1);
        xqo.setpart_ex(0, p, 2 * Np);

        fopr_ov5d->Mopr_5d_eo(p, x, 1);
        fopr_ov5d->LUprecond(t, p, 1);
        xt.setpart_ex(0, t, 2 * Np);
        xqo -= xt;

        index_eo.reverseField(xq, xqe, 0);
        index_eo.reverseField(xq, xqo, 1);

        fopr_ov->set_mode("H");
        v1  = fopr_ov->mult(xq);
        v1 -= (Field)b;
        vv  = v1 * v1;
        vv  = vv * snorm;
        vout.general(vl, "  Diff(ov)  = %16.8e\n", vv);

        sq[icolor + Nc * ispin] = (Field_F)xq;
      }
    }

    double result;

    CRSsolver(solution_file, matrix_file, source_file, result);


    // ####  tydy up  ####
    delete U;
    delete gconf_read;
    delete Ueo;

    delete fopr_crs;
    delete fopr_ov;
    delete fopr_w;

    delete fopr_ov5d;
    delete fopr_w_eo;

    delete solver;
    delete source;


#ifdef USE_TEST
    return Test::verify(expected_result, result);

#else
    return EXIT_SUCCESS;
#endif
  }
} // namespace Test_Spectrum_CRSMatrix
