/*!
        @file    $Id: test_Spectrum_Overlap_2ptFunction.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate: 2013-04-27 12:28:50 #$

        @version $LastChangedRevision: 934 $
*/

#include "parameterManager_YAML.h"
#include "parameters_factory.h"

#include "bridgeIO.h"
using Bridge::vout;

#include "gaugeConfig.h"
#include "staples.h"

#include "randomNumbers_Mseries.h"

#include "fopr_Wilson.h"
#include "fopr_Overlap.h"

#include "eigensolver_IRLanczos.h"

#include "gammaMatrixSet.h"
#include "gaugeFixing.h"
#include "solver.h"
#include "source.h"

#include "corr2pt_4spinor.h"

#ifdef USE_TEST
#include "test.h"
#endif

#ifdef USE_TESTMANAGER_AUTOREGISTER
#include "testManager.h"
#endif

//====================================================================
//! Test class of overlap fermion operators.

/*!
    This class provides tests of overlap fermion operator with and
    without subtraction of low-lying modes of the base fermion operator.
    Whether subtraction is performed or not is controlled in the code
    explicitly by commenting out some lines.
    As the base fermion operator, Wilson fermion operator is used.
                                          [13 Apr 2012 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.                  [14 Nov 2012 Y.Namekawa]
    Selectors are implemented.            [03 Mar 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
*/

namespace Test_Spectrum_Overlap {
  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_Spectrum_Overlap_Hadron2ptFunction.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_Spectrum_Overlap : public Parameters {
     public:
      Parameters_Test_Spectrum_Overlap()
      {
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int hadron_2ptFunction(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
    static const bool is_registered = TestManager::RegisterTest(
      "Spectrum.Overlap.Hadron2ptFunction",
      hadron_2ptFunction
      );
  }
#endif

  //====================================================================
  int hadron_2ptFunction(void)
  {
    // ####  parameter setup  ####
    int Ndim = CommonParameters::Ndim();
    int Nc   = CommonParameters::Nc();
    int Nd   = CommonParameters::Nd();
    int Nvol = CommonParameters::Nvol();

    Parameters *params_test      = new Parameters_Test_Spectrum_Overlap;
    Parameters *params_gfix      = ParametersFactory::New("GaugeFixing");
    Parameters *params_wilson    = ParametersFactory::New("Fopr.Wilson");
    Parameters *params_overlap   = ParametersFactory::New("Fopr.Overlap");
    Parameters *params_irlanczos = ParametersFactory::New("Eigensolver.IRLanczos");
    Parameters *params_solver    = ParametersFactory::New("Solver");
    Parameters *params_source    = ParametersFactory::New("Source");

    Parameters *params_all = new Parameters;

    params_all->Register_Parameters("Test_Spectrum_Overlap", params_test);
    params_all->Register_Parameters("GaugeFixing", params_gfix);
    params_all->Register_Parameters("Fopr_Wilson", params_wilson);
    params_all->Register_Parameters("Fopr_Overlap", params_overlap);
    params_all->Register_Parameters("Eigensolver", params_irlanczos);
    params_all->Register_Parameters("Solver", params_solver);
    params_all->Register_Parameters("Source", params_source);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_read = params_test->get_string("gauge_config_type_input");
    const string readfile       = params_test->get_string("config_filename_input");
    const string str_vlevel     = params_test->get_string("verbose_level");
#ifdef USE_TEST
    const double expected_result = params_test->get_double("expected_result");
#endif

    const string str_gfix_type   = params_gfix->get_string("gauge_fixing_type");
    const string str_gmset_type  = params_wilson->get_string("gamma_matrix_type");
    const string str_solver_type = params_solver->get_string("solver_type");
    const string str_source_type = params_source->get_string("source_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_read  = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile    = %s\n", readfile.c_str());
    vout.general(vl, "  vlevel      = %s\n", str_vlevel.c_str());
    vout.general(vl, "  gfix_type   = %s\n", str_gfix_type.c_str());
    vout.general(vl, "  gmset_type  = %s\n", str_gmset_type.c_str());
    vout.general(vl, "  solver_type = %s\n", str_solver_type.c_str());
    vout.general(vl, "  source_type = %s\n", str_source_type.c_str());

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_read);
    err += ParameterCheck::non_NULL(readfile);

    if (err) {
      vout.crucial(vl, "Test_Spectrum_Overlap: Input parameters have not been set.\n");
      abort();
    }


    // ####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);
    gconf_read->read_file((Field *)U, readfile);
    // gconf_read->set_cold((Field*)U);


    // ####  Gauge fixing  ####
    Staples *staple = new Staples;
    Field_G *Ufix   = new Field_G(Nvol, Ndim);

    int           ndelay = 1000;
    RandomNumbers *rand  = new RandomNumbers_Mseries(ndelay);

    GaugeFixing *gfix = GaugeFixing::New(str_gfix_type, rand);
    gfix->set_parameters(*params_gfix);

    double plaq = staple->plaquette(*U);
    vout.general(vl, "plaq(original) = %18.14f\n", plaq);

    gfix->fix(*Ufix, *U);

    double plaq2 = staple->plaquette(*Ufix);
    vout.general(vl, "plaq(fixed)    = %18.14f\n", plaq2);
    vout.general(vl, "plaq(diff)     = %18.10e\n", plaq - plaq2);


    // ####  object setup  ####
    GammaMatrixSet *gmset = GammaMatrixSet::New(str_gmset_type);

    Fopr_Wilson *fopr_w = new Fopr_Wilson(str_gmset_type);
    fopr_w->set_parameters(*params_wilson);
    fopr_w->set_config(Ufix);

    //-  with low-mode subtraction
    // int Nsbt = 0;
    // valarray<double> TDa;
    // valarray<Field>  vk;
    // calc_lowmodes(*params_irlanczos, Nsbt, TDa, vk, fopr_w);

    Fopr_Overlap *fopr_overlap = new Fopr_Overlap(fopr_w);
    fopr_overlap->set_parameters(*params_overlap);
    fopr_overlap->set_config(Ufix);
    // fopr_overlap->set_lowmodes(Nsbt, &TDa, &vk);

    Solver *solver = Solver::New(str_solver_type, fopr_overlap);
    solver->set_parameters(*params_solver);

    Source *source = Source::New(str_source_type);
    source->set_parameters(*params_source);


    // ####  Execution main part  ####
    valarray<Field_F> sq(Nc * Nd);
    Field_F           b;

    vout.general(vl, "\n");
    vout.general(vl, "Solving quark propagator:\n");
    vout.general(vl, "  color spin   Nconv      diff           diff2\n");

    // { int ispin = 0;
    //  { int icolor = 0;
    for (int ispin = 0; ispin < Nd; ++ispin) {
      for (int icolor = 0; icolor < Nc; ++icolor) {
        int idx = icolor + Nc * ispin;
        source->set(b, idx);

        int    Nconv;
        double diff;

        fopr_overlap->set_mode("D");
        solver->solve(sq[idx], b, Nconv, diff);

        Field y(b);
        fopr_overlap->set_mode("D");
        y -= fopr_overlap->mult(sq[idx]);
        double diff2 = y.norm2();

        vout.general(vl, "   %2d   %2d   %6d   %12.4e   %12.4e\n",
                     icolor, ispin, Nconv, diff, diff2);
      }
    }


    //- meson correlators
    std::ofstream log_file;
    if (filename_output != "stdout") {
      log_file.open(filename_output.c_str());
      vout.init(log_file);
    }

    vout.general(vl, "\n");
    vout.general(vl, "2-point correlator:\n");

    Corr2pt_4spinor corr(gmset);
    //corr.meson_all(sq,sq);
    double result = corr.meson_all(sq, sq);

    if (filename_output != "stdout") {
      log_file.close();
      vout.init(std::cout);
    }


    // ####  tidy up  ####
    delete gconf_read;
    delete U;
    delete Ufix;

    delete rand;
    delete staple;

    delete fopr_w;
    delete fopr_overlap;

    delete gfix;
    delete gmset;
    delete solver;
    delete source;

    delete params_test;
    delete params_gfix;
    delete params_wilson;
    delete params_overlap;
    delete params_irlanczos;
    delete params_solver;
    delete params_source;
    delete params_all;


#ifdef USE_TEST
    return Test::verify(expected_result, result);

#else
    return EXIT_SUCCESS;
#endif
  }
} // namespace Test_Spectrum_Overlap
