/*!
        @file    $Id:: test_Spectrum_Overlap_checkSignFunction.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-07-18 11:55:46 #$

        @version $LastChangedRevision: 934 $
*/

#include "parameterManager_YAML.h"
#include "parameters_factory.h"

#include "bridgeIO.h"
using Bridge::vout;

#include "gaugeConfig.h"
#include "staples.h"

#include "randomNumbers_Mseries.h"

#include "fopr_Wilson.h"
#include "fopr_Sign.h"

#include "gaugeFixing.h"
#include "source.h"

#include "corr2pt_4spinor.h"

#ifdef USE_TEST
#include "test.h"
#endif

#ifdef USE_TESTMANAGER_AUTOREGISTER
#include "testManager.h"
#endif

//====================================================================
//! Test class of sign function for overlap fermion operators.

/*!
    This class provides tests of sig function for overlap fermion
    operator.
                                          [13 Apr 2012 H.Matsufuru]
    (Coding history will be recovered from trac.)
    Selectors are implemented.            [03 Mar 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
 */

namespace Test_Spectrum_Overlap {
  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_Spectrum_Overlap_CheckSignFunction.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_Spectrum_Overlap : public Parameters {
     public:
      Parameters_Test_Spectrum_Overlap()
      {
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int check_sign(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
    static const bool is_registered = TestManager::RegisterTest(
      "Spectrum.Overlap.CheckSignFunction",
      check_sign
      );
  }
#endif

  //====================================================================
  int check_sign(void)
  {
    // ####  parameter setup  ####
    int Ndim = CommonParameters::Ndim();
    int Nc   = CommonParameters::Nc();
    int Nvol = CommonParameters::Nvol();

    Parameters *params_test    = new Parameters_Test_Spectrum_Overlap;
    Parameters *params_gfix    = ParametersFactory::New("GaugeFixing");
    Parameters *params_wilson  = ParametersFactory::New("Fopr.Wilson");
    Parameters *params_overlap = ParametersFactory::New("Fopr.Sign");
    Parameters *params_source  = ParametersFactory::New("Source");

    Parameters *params_all = new Parameters;

    params_all->Register_Parameters("Test_Spectrum_Overlap", params_test);
    params_all->Register_Parameters("GaugeFixing", params_gfix);
    params_all->Register_Parameters("Fopr_Wilson", params_wilson);
    params_all->Register_Parameters("Fopr_Overlap", params_overlap);
    params_all->Register_Parameters("Source", params_source);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_read = params_test->get_string("gauge_config_type_input");
    const string readfile       = params_test->get_string("config_filename_input");
    const string str_vlevel     = params_test->get_string("verbose_level");
#ifdef USE_TEST
    const double expected_result = params_test->get_double("expected_result");
#endif

    const string str_gfix_type   = params_gfix->get_string("gauge_fixing_type");
    const string str_gmset_type  = params_wilson->get_string("gamma_matrix_type");
    const string str_source_type = params_source->get_string("source_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_read  = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile    = %s\n", readfile.c_str());
    vout.general(vl, "  vlevel      = %s\n", str_vlevel.c_str());
    vout.general(vl, "  gfix_type   = %s\n", str_gfix_type.c_str());
    vout.general(vl, "  gmset_type  = %s\n", str_gmset_type.c_str());
    vout.general(vl, "  source_type = %s\n", str_source_type.c_str());

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_read);
    err += ParameterCheck::non_NULL(readfile);

    if (err) {
      vout.crucial(vl, "Test_Spectrum_Overlap: Input parameters have not been set.\n");
      abort();
    }


    // ####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);
    gconf_read->read_file((Field *)U, readfile);
    // gconf_read->set_cold((Field*)U);


    // ####  Gauge fixing  ####
    Staples *staple = new Staples;
    Field_G *Ufix   = new Field_G(Nvol, Ndim);

    int           ndelay = 1000;
    RandomNumbers *rand  = new RandomNumbers_Mseries(ndelay);

    GaugeFixing *gfix = GaugeFixing::New(str_gfix_type, rand);
    gfix->set_parameters(*params_gfix);

    double plaq = staple->plaquette(*U);
    vout.general(vl, "plaq(original) = %18.14f\n", plaq);

    gfix->fix(*Ufix, *U);

    double plaq2 = staple->plaquette(*Ufix);
    vout.general(vl, "plaq(fixed)    = %18.14f\n", plaq2);
    vout.general(vl, "plaq(diff)     = %18.10e\n", plaq - plaq2);


    // ####  object setup  ####
    Fopr_Wilson *fopr_w = new Fopr_Wilson(str_gmset_type);
    fopr_w->set_parameters(*params_wilson);
    fopr_w->set_config(Ufix);

    int              Nsbt = 0;
    valarray<double> TDa;
    valarray<Field>  vk;

    //- with low-mode subtraction
    //double Vthreshold = 0.15;  // example value with low-mode subtraction
    //calc_lowmodes(Nsbt, TDa, vk, Vthreshold, fopr_w);

    Fopr_Sign *fopr_sign = new Fopr_Sign(fopr_w);
    fopr_sign->set_parameters(*params_overlap);
    fopr_sign->set_config(Ufix);
    if (Nsbt > 0) fopr_sign->set_lowmodes(Nsbt, &TDa, &vk);

    Source *source = Source::New(str_source_type);
    source->set_parameters(*params_source);


    // ####  Execution main part  ####
    Field_F b;

    Field xq((Field)b);
    Field xq2((Field)b);

    double xq_norm;
    {
      int ispin = 0;
      {
        int icolor = 0;
        //  for(int ispin = 0; ispin < Nd; ++ispin){
        //    for(int icolor = 0; icolor < Nc; ++icolor){

        int idx = icolor + Nc * ispin;
        source->set(b, idx);

        xq   = fopr_sign->mult((Field)b);
        xq2  = fopr_sign->mult(xq);
        xq2 -= (Field)b;

        xq_norm = xq2.norm2();

        vout.general(vl, " |sign^2(b) - b|^2 = %16.8e\n", xq_norm);
      }
    }

    double result = xq_norm;


    // #####  tidy up  #####
    delete gconf_read;
    delete U;
    delete Ufix;

    delete rand;
    delete staple;

    delete fopr_w;
    delete fopr_sign;

    delete gfix;
    delete source;

    delete params_test;
    delete params_gfix;
    delete params_wilson;
    delete params_overlap;
    delete params_source;
    delete params_all;


#ifdef USE_TEST
    return Test::verify(expected_result, result);

#else
    return EXIT_SUCCESS;
#endif
  }
} // namespace Test_Spectrum_Overlap
