/*!
        @file    $Id:: test_Spectrum_Staggered_2ptFunction_eo_WallSource.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2013-04-27 12:28:50 #$

        @version $LastChangedRevision: 875 $
*/

#include "parameterManager_YAML.h"
#include "parameters_factory.h"

#include "bridgeIO.h"
using Bridge::vout;

#include "gaugeConfig.h"
#include "staples.h"

#include "randomNumbers_Mseries.h"

#include "fopr_Staggered_eo.h"
#include "fprop_Staggered_eo.h"
#include "source_Staggered_Wall.h"

#include "gaugeFixing.h"

#include "corr2pt_Staggered.h"

#ifdef USE_TEST
#include "test.h"
#endif

#ifdef USE_TESTMANAGER_AUTOREGISTER
#include "testManager.h"
#endif

//====================================================================
//! Spectrum test of staggered fermion.

/*!
    Temporary implementation.
                                          [28 Dec 2011 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.                  [14 Nov 2012 Y.Namekawa]
    Selectors are implemented.            [03 Mar 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
 */

namespace Test_Spectrum_Staggered {
  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_Spectrum_Staggered_Hadron2ptFunction_WallSource.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_Spectrum_Staggered : public Parameters {
     public:
      Parameters_Test_Spectrum_Staggered()
      {
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int hadron_2ptFunction_eo_wallSource(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
    static const bool is_registered = TestManager::RegisterTest(
      "Spectrum.Staggered.Hadron2ptFunction_eo.WallSource",
      hadron_2ptFunction_eo_wallSource
      );
  }
#endif

  //====================================================================
  int hadron_2ptFunction_eo_wallSource(void)
  {
    // ####  parameter setup  ####
    int Nc   = CommonParameters::Nc();
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();

    Parameters *params_test      = new Parameters_Test_Spectrum_Staggered;
    Parameters *params_gfix      = ParametersFactory::New("GaugeFixing");
    Parameters *params_staggered = ParametersFactory::New("Fopr.Staggered_eo");
    Parameters *params_solver    = ParametersFactory::New("Solver");
    Parameters *params_source    = ParametersFactory::New("Source_Staggered_Wall");

    Parameters *params_all = new Parameters;

    params_all->Register_Parameters("Test_Spectrum_Staggered", params_test);
    params_all->Register_Parameters("GaugeFixing", params_gfix);
    params_all->Register_Parameters("Fopr_Staggered", params_staggered);
    params_all->Register_Parameters("Solver", params_solver);
    params_all->Register_Parameters("Source_Wall", params_source);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_read = params_test->get_string("gauge_config_type_input");
    const string readfile       = params_test->get_string("config_filename_input");
    const string str_vlevel     = params_test->get_string("verbose_level");
#ifdef USE_TEST
    const double expected_result = params_test->get_double("expected_result");
#endif

    const string str_gfix_type = params_gfix->get_string("gauge_fixing_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_read = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile   = %s\n", readfile.c_str());
    vout.general(vl, "  vlevel     = %s\n", str_vlevel.c_str());
    vout.general(vl, "  gfix_type  = %s\n", str_gfix_type.c_str());

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_read);
    err += ParameterCheck::non_NULL(readfile);

    if (err) {
      vout.crucial(vl, "Test_Spectrum_Staggered: Input parameters have not been set.\n");
      abort();
    }


    // ####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);
    gconf_read->read_file((Field *)U, readfile);
    // gconf_read->set_cold((Field*)U);


    // ####  Gauge fixing  ####
    Staples *staple = new Staples;
    Field_G *Ufix   = new Field_G(Nvol, Ndim);

    int           ndelay = 1000;
    RandomNumbers *rand  = new RandomNumbers_Mseries(ndelay);

    GaugeFixing *gfix = GaugeFixing::New(str_gfix_type, rand);
    gfix->set_parameters(*params_gfix);

    double plaq = staple->plaquette(*U);
    vout.general(vl, "plaq(original) = %18.14f\n", plaq);

    gfix->fix(*Ufix, *U);

    double plaq2 = staple->plaquette(*Ufix);
    vout.general(vl, "plaq(fixed)    = %18.14f\n", plaq2);
    vout.general(vl, "plaq(diff)     = %18.10e\n", plaq - plaq2);


    Field_G  *Ueo = new Field_G(Nvol, Ndim);
    Index_eo index_eo;
    index_eo.convertField(*Ueo, *Ufix);

    vout.general(vl, "m_conf norm2 = %20.16g\n", Ufix->norm2());
    vout.general(vl, "Ueo    norm2 = %20.16g\n", Ueo->norm2());


    // ####  object setup  ####
    Fopr_Staggered_eo *fopr_ks = new Fopr_Staggered_eo;
    fopr_ks->set_parameters(*params_staggered);
    fopr_ks->set_config((Field *)Ueo);

    Fprop_Staggered_eo fprop_eo(fopr_ks);
    fprop_eo.set_parameters(*params_solver);

    Source_Staggered_Wall *source = new Source_Staggered_Wall();
    source->set_parameters(*params_source);


    // ####  Execution main part  ####
    Field_F_1spinor           src;
    valarray<Field_F_1spinor> sq(Nc * 2);

    for (int isrc = 0; isrc < 2; ++isrc) {
      for (int ic = 0; ic < Nc; ++ic) {
        vout.general(vl, "ic = %d, isrc = %d\n", ic, isrc);

        source->set(src, ic, isrc);

        int idx = ic + Nc * isrc;
        fprop_eo.calc(sq[idx], src);
      }
    }


    //- meson correlators
    vout.general(vl, "\n");
    vout.general(vl, "meson correlator:\n");

    Corr2pt_Staggered corr2pt;
    valarray<double>  mcorr;
    corr2pt.meson(mcorr, sq, sq);

    vout.general(vl, "PS <-- PS correlator:\n");
    for (int t = 0; t < mcorr.size(); ++t) {
      vout.general(vl, "  %4d %20.12e\n", t, mcorr[t]);
    }

    double result = mcorr[0];


    // #####  tidy up  #####
    delete gconf_read;
    delete U;
    delete Ueo;

    delete fopr_ks;
    delete source;

    delete params_test;
    delete params_gfix;
    delete params_staggered;
    delete params_solver;
    delete params_source;
    delete params_all;


#ifdef USE_TEST
    return Test::verify(expected_result, result);

#else
    return EXIT_SUCCESS;
#endif
  }
} // namespace Test_Spectrum_Staggered
